'''
from bioservices.services import REST


__all__ = ["PubChem"]


class PubChem(REST):
    """Interface to the `PubChem <todo>`_ service

    """
    _url = "http://pubchem.ncbi.nlm.nih.gov/rest/pug"

    def __init__(self, verbose=False, cache=False):
        """**Constructor**

        :param verbose: set to False to prevent informative messages
        """
        print("PubChem is not finalised yet. This is currently only a draft version")
        super(PubChem, self).__init__(name="PubChem", url=PubChem._url,
                verbose=verbose, cache=cache)

    def get_compound_by_smiles(self, identifier, frmt='json'):

        res = self.http_get("compound/smiles/" + identifier
                + '/cids/%s' % frmt, frmt=frmt,
                headers=self.get_headers(content=frmt))
        return res

    def get_compound_by_name(self, name, frmt='json'):
        res = self.http_get("compound/name/{}/{}".format(name,frmt),
                            frmt=frmt,
                            headers=self.get_headers(content=frmt))

        return res

    def get_drug_classification(self, name, frmt='json'):
        res = self.http_get("compound/name/{}/classification/{}".format(name, frmt),
                            frmt=frmt,
                            headers=self.get_headers(content=frmt))
        return res


def test_drug_classification():
    pc = PubChem()
    x = pc.get_drug_classification('cis-Diaminedichloroplatinum')
    print(x.keys())
    hie = x['Hierarchies']['Hierarchy']
    for i in hie:
        print(i['SourceName'])
        print(i.keys())
        if i['SourceName'] == 'MeSH':
            for n in i['Node']:
                print(n)
                print(n['Information']['Name'])
                if 'Description' in n['Information']:
                    print(n['Information']['Description'])
                if n['Information']['Name'] == 'Molecular Mechanisms of Pharmacological Action':
                    print(n)

        # print(i.keys())
        # for k in i.keys():
        #     print(k)




def test_drug_chem_properties():
    pc = PubChem()
    x = pc.get_compound_by_name('Tazobactam')
    dtype_dict = dict()
    dtype_dict[1] = 'sval'
    dtype_dict[7] = 'fval'
    dtype_dict[5] = 'ival'
    dtype_dict[16] = 'binary'

    for i in x['PC_Compounds']:
        print(i)
        print(i.keys())
        # [u'count', u'stereo', u'bonds', u'atoms', u'charge', u'coords', u'props', u'id']
        keys = ['name']
        for j in i['props']:
            # print(j)
            print(j['value'][dtype_dict[j['urn']['datatype']]])
            print()
            continue

            if 'name' in j['urn']:
                print(j['urn']['name'])

            if 'label' in j['urn']:
                print(j['urn']['label'])
                print(j['urn'], j['value'])


if __name__ == '__main__':

    # test_drug_chem_properties()
    test_drug_classification()
    quit()
    pc = PubChem()
    # x = pc.get_compound_by_smiles('CC1(C(N2C(S1(=O)=O)CC2=O)C(=O)O)CN3C=CN=N3')
    # print(x)

    # x = pc.get_compound_by_name('Tazobactam')
'''



import io
import logging
import os
import urllib.request

import networkx as nx
import pandas as pd

from magine.data.storage import network_data_dir
from magine.logging import get_logger

_p_name = os.path.join(network_data_dir, 'reactome_fi.p.gz')
logger = get_logger(__name__, log_level=logging.INFO)


def load_reactome_fi():
    """
    Load reactome functional interaction network

    Returns
    -------
    pandas.DataFrame
    """

    if not os.path.exists(_p_name):
        download_reactome_fi()
        if not os.path.exists(_p_name):
            raise FileNotFoundError("Error downloading reactome FI. ")
    tmp_graph = nx.read_gpickle(_p_name)

    n_n, n_e = len(tmp_graph.nodes()), len(tmp_graph.edges())
    logger.info("Reactome : {} nodes and {} edges".format(n_n, n_e))
    return tmp_graph


_maps = {

    'inhibited': 'inhibit',
    'inhibite': 'inhibit',
    'inhibition': 'inhibit',

    'expression': 'expression',
    'expressed': 'expression',

    'repressed': 'repression',

    'catalyzed': 'catalyze',

    'complex': 'binding',
    'dissociation': 'binding',

    'activated': 'activate',
    'activation': 'activate',

    'phosphorylated': 'phosphorylate',
    'phosphorylation': 'phosphorylate',

    'dephosphorylation': 'dephosphorylate',
    'dephosphorylated': 'dephosphorylate',

    'ubiquitinated': 'ubiquitinate',
    'ubiquitination': 'ubiquitinate',
    'glycosylation': 'glycosylate',
    'methylation': 'methylate',

    'binding/association': 'binding',

}


def standardize_edge_types(row):
    name = row['Annotation']
    name = name.replace(':', ' ').replace(';', ' ').replace(',', ' ')
    name = name.replace('state change', 'stateChange')
    name = set(name.split(' '))
    to_remove = ['by', 'regulates', 'regulated', 'input',
                 'PPrel', 'PCrel', 'GErel', 'ECrel', 'interaction',
                 '',
                 ]
    for i in to_remove:
        if i in name:
            name.remove(i)
    for k, v in _maps.items():
        if k in name:
            name.remove(k)
            name.add(v)
    name = '|'.join(sorted(name))
    return name


def download_reactome_fi():
    """ Downloads reactome functional interaction network """

    logger.info("Downloading Reactome Functional interaction network")

    url = 'http://cpws.reactome.org/caBigR3WebApp2019/FIsInGene_020720_with_annotations.txt.zip'

    req = urllib.request.Request(url)
    with urllib.request.urlopen(req) as response:
        # the_page = response.read()
        table = pd.read_csv(
            io.BytesIO(response.read()), compression='zip', delimiter='\t',
            error_bad_lines=False, encoding='utf-8'
        )
    table = table[table['Direction'] != '-']
    table = table[~table['Annotation'].str.contains('indirect effect')]
    table = table[~table['Annotation'].str.contains('predicted')]
    table = table[~table['Annotation'].str.contains('compound')]
    genes = set(table['Gene1'])
    genes.update(set(table['Gene2']))
    from magine.mappings.gene_mapper import GeneMapper
    gm = GeneMapper()
    missing_uniprot = set(i for i in genes if i not in gm.gene_name_to_uniprot)
    table = table[~table['Gene1'].isin(missing_uniprot)]
    table = table[~table['Gene2'].isin(missing_uniprot)]

    table['source'] = table['Gene1']
    table['target'] = table['Gene2']
    table['databaseSource'] = 'ReactomeFI'
    rev_cols = table['Direction'].isin(_reverse)

    table.loc[rev_cols, ['source', 'target']] = \
        table.loc[rev_cols, ['target', 'source']].values

    table['interactionType'] = table.apply(standardize_edge_types, axis=1)
    protein_graph = nx.from_pandas_edgelist(
        table,
        'source',
        'target',
        edge_attr=['interactionType', 'databaseSource'],
        create_using=nx.DiGraph()
    )
    species = set(table['source'].unique()).union(
        set(table['target'].unique())
    )

    # add names to graph
    for node in species:
        protein_graph.add_node(node, databaseSource='ReactomeFI',
                               speciesType='gene')

    logger.info("Done downloading Reactome functional interaction network")
    nx.write_gpickle(protein_graph, _p_name)


_reverse = {"<-", "|-"}
_forward = {"->", "->"}
_both = {'<->', '<-|', '|->', '|-|'}


if __name__ == '__main__':
    download_reactome_fi()



"""

import os
import networkx as nx
from magine.networks.visualization.cytoscape_view import RenderModel


def test_by_time():
    # return
    # This test must be ran with a session of cytoscape open
    # For now this will be passed in the travis builds.
    dir_name = os.path.dirname(__file__)
    full_path = os.path.join(dir_name,
                             'Network_files',
                             'test_all_colored_enrichment.graphml')
    ddn = nx.nx.read_graphml(full_path)
    rm = RenderModel(ddn, style='Directed', )  # layout='force-directed')
    rm.visualize_by_list_of_time(
        ['time_0000', 'time_0001', 'time_0002', 'time_0003'],
        labels=[0, 2, 3, 4], out_dir='tmp')
    files_that_should_exist = [
        'tmp/Figures/ont_network_tmp_time_0000_formatted.png',
        'tmp/Figures/ont_network_tmp_time_0001_formatted.png',
        'tmp/Figures/ont_network_tmp_time_0002_formatted.png',
        'tmp/Figures/ont_network_tmp_time_0003_formatted.png',
    ]
    for i in files_that_should_exist:
        assert os.path.exists(i)


if __name__ == '__main__':
    test_by_time()
"""



import json
import logging
import os
import re

# Will be OK in Python 2
try:
    basestring
# Allows isinstance(foo, basestring) to work in Python 3
except:
    basestring = str

import numpy as np
import pandas as pd
import requests

from magine.plotting.species_plotting import write_table_to_html
from magine.enrichment.enrichment_result import EnrichmentResult
from magine.logging import get_logger

logger = get_logger(__name__, log_level=logging.INFO)

_valid_libs = set()
with open(os.path.join(os.path.dirname(__file__),
                       '_valid_enricher_libs.txt'), 'r') as f:
    for n in f.read().splitlines():
        _valid_libs.add(n)

gene = 'gene'

db_types = {
    'histone': [
        'Epigenomics_Roadmap_HM_ChIP-seq',
        'ENCODE_Histone_Modifications_2015',
        'ESCAPE',
    ],
    'mrna': [
        'TargetScan_microRNA_2017',
        'miRTarBase_2017',
    ],
    'kinases': [
        'KEA_2015',
        'Kinase_Perturbations_from_GEO_down',
        'Kinase_Perturbations_from_GEO_up',
        'Phosphatase_Substrates_from_DEPOD',
        'ARCHS4_Kinases_Coexp',
        'ARCHS4_IDG_Coexp',
    ],
    'transcription': [
        'ChEA_2016',
        'TRANSFAC_and_JASPAR_PWMs',
        'ARCHS4_TFs_Coexp',
        'Enrichr_Submissions_TF-Gene_Coocurrence',
        'Genome_Browser_PWMs',
        'ENCODE_and_ChEA_Consensus_TFs_from_ChIP-X',
        'ENCODE_TF_ChIP-seq_2015',
        'TF-LOF_Expression_from_GEO',
        'Transcription_Factor_PPIs'
    ],
    'pathways': [
        'KEGG_2016',
        'WikiPathways_2016',
        'Reactome_2016',
        'BioCarta_2016',
        'NCI-Nature_2016',
        'Panther_2016',
    ],
    'complex': [
        'NURSA_Human_Endogenous_Complexome',
        'CORUM',
        'PPI_Hub_Proteins',
        'BioPlex_2017',
    ],
    'ontologies': [
        'GO_Cellular_Component_2018',
        'GO_Biological_Process_2018',
        'GO_Molecular_Function_2018',
        'MGI_Mammalian_Phenotype_2017',
        'Jensen_COMPARTMENTS',
    ],
    'drug': [
        'DrugMatrix',
        'Drug_Perturbations_from_GEO_2014',
        'Old_CMAP_up',
        'Old_CMAP_down',
        'LINCS_L1000_Chem_Pert_up',
        'LINCS_L1000_Chem_Pert_down',
        'LINCS_L1000_Ligand_Perturbations_up',
        'LINCS_L1000_Ligand_Perturbations_down',
    ],
    'disease': [
        'OMIM_Disease',
        'OMIM_Expanded',
        'Jensen_DISEASES',
        'Human_Phenotype_Ontology'
    ],
    'cell_type': [

        'ARCHS4_Tissues',
        'ARCHS4_Cell-lines',
        'Allen_Brain_Atlas_up',
        'Allen_Brain_Atlas_down',
        'Cancer_Cell_Line_Encyclopedia',
        'Human_Gene_Atlas',
        'NCI-60_Cancer_Cell_Lines',
        'Jensen_TISSUES',
    ],
    'crowd': [
        'Disease_Perturbations_from_GEO_down',
        'Disease_Perturbations_from_GEO_up',
        'Drug_Perturbations_from_GEO_down',
        'Drug_Perturbations_from_GEO_up',
        'RNA-Seq_Disease_Gene_and_Drug_Signatures_from_GEO',
        'Aging_Perturbations_from_GEO_up',
        'Aging_Perturbations_from_GEO_down',
        'Ligand_Perturbations_from_GEO_up',
        'Ligand_Perturbations_from_GEO_down',
        'MCF7_Perturbations_from_GEO_up',
        'MCF7_Perturbations_from_GEO_down',
        'Microbe_Perturbations_from_GEO_up',
        'Microbe_Perturbations_from_GEO_down',
        'SysMyo_Muscle_Gene_Sets',
    ]

}


def get_libraries():
    url = 'http://amp.pharm.mssm.edu/Enrichr/datasetStatistics'
    libs_json = json.loads(requests.get(url).text)
    return [lib['libraryName'] for lib in libs_json['statistics']]


class Enrichr(object):
    _query = '{url}/enrich?userListId={list_id}&backgroundType={lib}'

    def __init__(self, verbose=False):
        self._url = 'http://amp.pharm.mssm.edu/Enrichr'
        self._valid_libs = _valid_libs
        self.verbose = verbose

    def print_valid_libs(self):
        """
        Print a list of all available libraries EnrichR has to offer.



        """
        for lib_name in sorted(self._valid_libs):
            print(lib_name)

    def run(self, list_of_genes, gene_set_lib='GO_Biological_Process_2017'):
        """

        Parameters
        ----------
        list_of_genes : list_like
            List of genes using HGNC gene names
        gene_set_lib : str or list
            Name of gene set library
            To print options use Enrichr.print_valid_libs


        Examples
        --------
        >>> import pandas as pd
        >>> pd.set_option('display.max_colwidth', 40)
        >>> pd.set_option('precision', 3)
        >>> e = Enrichr()
        >>> df = e.run(['BAX', 'BCL2', 'CASP3', 'CASP8'], \
                        gene_set_lib='Reactome_2016')
        >>> print(df[['term_name','combined_score']].head(5))#doctest: +NORMALIZE_WHITESPACE
                                             term_name  combined_score
            0  intrinsic pathway for apoptosis hsa ...       11814.410
            1               apoptosis hsa r-hsa-109581        2365.141
            2  programmed cell death hsa r-hsa-5357801        2313.527
            3  caspase-mediated cleavage of cytoske...       10944.261
            4  caspase activation via extrinsic apo...        4245.542

        Returns
        -------
        df : EnrichmentResult
            Results from enrichR

        """
        if not isinstance(list_of_genes, (list, set)):
            raise AssertionError("list_of_genes must be list like")
        logger.debug("Running Enrichr with gene set {}".format(gene_set_lib))
        user_list_id = self._add_gene_list(list_of_genes)
        if isinstance(gene_set_lib, basestring):
            df = self._run_id(user_list_id, gene_set_lib)
        else:
            df = self._run_list_of_dbs(user_list_id, gene_set_lib)

        init_size = len(df)
        if init_size == 0:
            logger.debug("No terms returned")
            return df
        try:
            df = _prepare_output(df)
        except:
            logger.warning("Unable to simplify enrichR term names. This could "
                           "be due to a change in enrichR formatting. "
                           "Returning as default output.")

        # set all terms with adj_p_value less than 0.05 to significant.
        df['significant'] = False
        crit = (df['adj_p_value'] <= 0.05)
        df.loc[crit, 'significant'] = True

        after_size = len(df)
        if init_size != after_size:
            raise AssertionError('not the same shape {}'.format(gene_set_lib))
        logger.debug("Done calling Enrichr.")
        return df

    def run_samples(self, sample_lists, sample_ids,
                    gene_set_lib='GO_Biological_Process_2017', save_name=None,
                    create_html=False, out_dir=None, run_parallel=False,
                    exp_data=None, pivot=False):
        """ Run enrichment analysis on a list of samples.

        Parameters
        ----------
        sample_lists : list_like
            List of lists of genes for enrichment analysis
        sample_ids : list
            list of ids for the provided sample list
        gene_set_lib : str, list
            Type of gene set, refer to Enrichr.print_valid_libs
        save_name : str, optional
            if provided it will save a file as a pivoted table with
            the term_ids vs sample_ids
        create_html : bool
            Creates html of output with plots of species across sample
        out_dir : str
            If create_html, it will place all html plots into this directory
        run_parallel : bool
            If create_html, it will create plots using multiprocessing
        exp_data : magine.data.ExperimentalData
            Must be provided if create_html=True
        pivot : bool

        Examples
        --------
        .. plot::
            :context: close-figs

            >>> import pandas as pd
            >>> import matplotlib.pyplot as plt
            >>> from magine.enrichment.enrichr import Enrichr
            >>> pd.set_option('display.max_colwidth', 40)
            >>> pd.set_option('precision', 3)
            >>> samples = [['BAX', 'BCL2', 'CASP3', 'CASP8'], \
['ATR', 'ATM', 'TP53', 'CHEK1']]
            >>> sample_ids = ['apoptosis', 'dna_repair']
            >>> e = Enrichr()
            >>> df = e.run_samples(samples, sample_ids, \
gene_set_lib='Reactome_2016')
            >>> print(df[['term_name','combined_score']].head(5))#doctest: +NORMALIZE_WHITESPACE
                                                 term_name  combined_score
                0  intrinsic pathway for apoptosis hsa ...       11814.410
                1               apoptosis hsa r-hsa-109581        2365.141
                2  programmed cell death hsa r-hsa-5357801        2313.527
                3  caspase-mediated cleavage of cytoske...       10944.261
                4  caspase activation via extrinsic apo...        4245.542
            >>> df.filter_multi(rank=10, inplace=True)
            >>> df['term_name'] = df['term_name'].str.split('_').str.get(0)
            >>> fig = df.sig.heatmap(figsize=(6, 6), linewidths=.05)

        Returns
        -------
        EnrichmentResult
        """
        if not isinstance(sample_lists, list):
            raise AssertionError("List required")
        if not isinstance(sample_lists[0], (list, set)):
            raise AssertionError("List of lists required")
        df_final = self.run(sample_lists[0], gene_set_lib)
        df_final['sample_id'] = sample_ids[0]
        for count, (i, j) in enumerate(zip(sample_lists[1:], sample_ids[1:])):
            df = self.run(i, gene_set_lib)
            df['sample_id'] = j
            df_final = df_final.append(df, ignore_index=True)

        # removes terms that do not have at least 1 signficant term across any
        # of the samples list provided
        df_final.require_n_sig(n_sig=1, inplace=True)
        df_final = EnrichmentResult(df_final)
        if save_name:
            s_name = '{}_enrichr'.format(save_name)
            if pivot:
                # pivot output
                p_df = pd.pivot_table(
                    df_final, index=['term_name', 'db'],
                    columns='sample_id',
                    aggfunc='first',
                    fill_value=np.nan,
                    values=['term_name', 'rank', 'p_value', 'z_score',
                            'combined_score', 'adj_p_value', 'genes',
                            'significant'],
                )
                # save files
                p_df.to_excel('{}.xlsx'.format(s_name), merge_cells=True)
            df_final.to_csv('{}.csv'.format(s_name), index=False)

        if create_html:
            if exp_data is None:
                raise AssertionError("exp_data required for plots")

            write_table_to_html(data=df_final, save_name=save_name,
                                out_dir=out_dir, run_parallel=run_parallel,
                                exp_data=exp_data)
        return df_final

    def _run_id(self, list_id, gene_set_lib):
        if gene_set_lib not in _valid_libs:
            raise AssertionError("{} not in valid ids {}".format(gene_set_lib,
                                                                 _valid_libs))

        q = self._query.format(url=self._url, list_id=list_id,
                               lib=gene_set_lib)
        response = requests.get(q, timeout=100)

        if not response.ok:
            raise Exception("EnrichR response error")
        data = json.loads(response.text)
        if gene_set_lib not in data:
            raise Exception("{} not in enrichR".format(gene_set_lib))
        if len(data[gene_set_lib]) == 0:
            return EnrichmentResult()
        #####
        # ENRICHR return a list of entries with each entry having these terms
        # Rank, Term name, P-value, Z-score, Combined score, Overlapping genes,
        # Adjusted p-value, Old p-value, Old adjusted p-value
        #####

        df = EnrichmentResult(
            data[gene_set_lib],
            columns=['rank', 'term_name', 'p_value', 'z_score',
                     'combined_score', 'gene_hits', 'adj_p_value', '_', '_']
        )

        def compress_genes(row):
            return ','.join(g for g in sorted(row['gene_hits']))

        def get_length(row):
            return len(row['gene_hits'])

        # compress genes into string and get count
        df['genes'] = df.apply(compress_genes, axis=1)
        df['n_genes'] = df.apply(get_length, axis=1)

        # remove term names if null
        df = df[~df['term_name'].isnull()].copy()

        # keep only desired columns
        cols = ['term_name', 'rank', 'p_value', 'z_score', 'combined_score',
                'adj_p_value', 'genes', 'n_genes']

        df = df[cols]
        df['db'] = gene_set_lib
        return df

    def _run_list_of_dbs(self, gene_list_id, databases):
        data = self._run_id(gene_list_id, databases[0])
        for db in databases[1:]:
            data = data.append(self._run_id(gene_list_id, db),
                               ignore_index=True)
            logger.debug('\t\t{}/{} databases'.format(db, len(databases)))
        return data

    def _add_gene_list(self, gene_list):
        """ Upload to enrichr

        Parameters
        ----------
        gene_list : list

        Returns
        -------
        str : userListId that maps to gene list provided, used within enrichR
        to gather enrichment results.
        """

        genes_str = '\n'.join(gene_list)

        payload = {
            'list': (None, genes_str),
            'description': (None, 'MAGINE analysis')
        }
        response = requests.post(self._url + '/addList', files=payload)
        if not response.ok:
            raise Exception('Error analyzing gene list', response.ok)

        data = json.loads(response.text)
        return data['userListId']


def _prepare_output(df):
    df['term_name'] = df.apply(clean_term_names, axis=1)

    return df


def clean_term_names(row):
    term_name = row['term_name']
    if not isinstance(term_name, basestring):
        return term_name

    db = row['db']

    if db in [
        'GO_Biological_Process_2018', 'GO_Molecular_Function_2018',
        'GO_Cellular_Component_2018',
        'GO_Biological_Process_2017', 'GO_Molecular_Function_2017',
        'GO_Cellular_Component_2017',
        'GO_Biological_Process_2017b', 'GO_Molecular_Function_2017b',
        'GO_Cellular_Component_2017b',
    ]:

        if 'GO:' in term_name:
            term_name = term_name.split('(GO:', 1)[0]
        elif 'go:' in term_name:
            term_name = term_name.split('(go:', 1)[0]
    if db == 'Human_Phenotype_Ontology':
        if 'HP:' in term_name:
            term_name = term_name.split('(HP:', 1)[0]

    if db == 'MGI_Mammalian_Phenotype_2017':
        if term_name.startswith('MP:'):
            term_name = term_name.split(' ', 1)[1]

    if db == 'DrugMatrix':
        drug_name = re.search(r'^(.*)(-\d*.*\d_)', term_name).group(1)
        direction = re.search(r'-(.{2})$', term_name).group(0)
        term_name = drug_name + direction

    if db in ['Old_CMAP_down', 'Old_CMAP_up']:
        term_name = term_name.rsplit('-', 1)[0]

    if db in ['Ligand_Perturbations_from_GEO_down',
              'Ligand_Perturbations_from_GEO_up']:
        term_name = term_name.split('_', 1)[0]

    term_name = term_name.strip()
    term_name = term_name.lower()
    for i, j in replace_pairs:
        if i in term_name:
            term_name = term_name.replace(i, j)

    return term_name


def clean_lincs(df):
    """ Cleans the lincs databases term_names from enrichR.

    Parameters
    ----------
    df

    Returns
    -------

    """
    pattern = re.compile(
        r'(?P<id>\w+) (?P<cell>\w+) (?P<time>\w+)-(?P<drug>.*)-(\S*)')

    def get_drug(row):
        db = row['db']
        term_name = row['term_name']
        if db not in ['LINCS_L1000_Chem_Pert_up',
                      'LINCS_L1000_Chem_Pert_down']:
            return term_name
        try:
            drug_name = pattern.search(term_name).group('drug')
            return drug_name
        except:
            # If the string doesnt have the pattern above, there is no way
            # to properly parse it. Ran into a few examples where they used
            # spaces instead of "_". Hard to parse since the names of the drugs
            # can also have spaces ("sulfide salts")
            # ex. cpc006 snuc5 6h-quinine hemisulfate salt monohydrate-10.0
            return term_name

    df['term_name'] = df.apply(get_drug, axis=1)
    return df


def clean_drug_pert_geo(df):
    def get_drug(row):
        db = row['db']
        term_name = row['term_name']
        if db != 'Drug_Perturbations_from_GEO_2014':
            return term_name
        try:
            segs = term_name.split('_')
            drug_name = segs[0] + '_' + segs[-1][0] + segs[-1][-1]
        except NameError:
            drug_name = term_name
        return drug_name

    df['term_name'] = df.apply(get_drug, axis=1)
    return df


def clean_drug_dbs(data):
    df_copy = data.copy()
    if 'Drug_Perturbations_from_GEO_2014' in df_copy.db.unique():
        df_copy = clean_drug_pert_geo(df_copy)
    if 'LINCS_L1000_Chem_Pert_up' in df_copy.db.unique() or \
            'LINCS_L1000_Chem_Pert_down' in df_copy.db.unique():
        df_copy = clean_lincs(df_copy)
    return df_copy


def clean_tf_names(data):
    """
    Cleans transcription factors databases by removing everything after '_'.

    Parameters
    ----------
    data : pd.DataFrame

    Returns
    -------

    """

    def return_single_tf(row):
        """
        Gets TF name only
        """
        tf = row['term_name']
        if '_' in tf:
            tf = tf.split('_')[0]
        return tf.upper()

    tf_dbs = ['ARCHS4_TFs_Coexp',
              'ChEA_2016',
              'ENCODE_and_ChEA_Consensus_TFs_from_ChIP-X',
              'ENCODE_TF_ChIP-seq_2015',
              'Enrichr_Submissions_TF-Gene_Coocurrence',
              'TRANSFAC_and_JASPAR_PWMs',
              'TF-LOF_Expression_from_GEO',
              'Transcription_Factor_PPIs']
    tfs = data[data['db'].isin(tf_dbs)].copy()
    tfs['term_name'] = tfs.apply(return_single_tf, axis=1)
    return tfs


replace_pairs = [
    ('mus musculus', 'mus'),
    ('homo sapiens', 'hsa'),
    ('rattus norvegicus', 'rat'),
    ('oncorhynchus mykiss', 'onc_mykiss'),
    ('macaca fascicularis', 'mfa'),
    ('oryzias latipes', 'ola'),
    ('sus scrofa', 'ssc'),
    ('danio rerio', 'dre'),
    ('bos taurus', 'bta'),
    ('dictyostelium discoideum', 'dicty'),
    ('myzus persicae', 'm.persicae'),
    ('staphylococcus aureus', 's.aureus'),
    ('escherichia coli', 'e.coli'),
    ('pseudomonas aeruginosa', 'p.aeruginosa'),
    ('drosophila melanogaste', 'fly'),
    ('mycobacterium tuberculosis', 'm.tuberculosis'),
    ('hordeum vulgare', 'h.vulgare'),
    (' (mouse)', '_mus'),
    (' (human)', '_hsa'),
    ('Homo sapiens', 'hsa'),
    ('Mus musculus', 'mus'),
    ('hg19', '_hsa'),
    ('mm9', '_mus'),
    (' hsa r-hsa', '_hsa')

]

standard_dbs = []
for db in ['drug', 'disease', 'ontologies', 'pathways', 'transcription',
           'kinases', 'histone', 'cell_type']:
    standard_dbs += db_types[db]


def get_background_list(lib_name):
    """
    Return reference list for given gene referecen set

    Parameters
    ----------
    lib_name : str

    Returns
    -------

    """
    enrichment_url = 'http://amp.pharm.mssm.edu/Enrichr/geneSetLibrary'
    query_string = '?userListId&libraryName={}'
    response = requests.get(
        enrichment_url + query_string.format(lib_name)
    )
    if not response.ok:
        raise Exception('Error fetching enrichment results')

    results = json.loads(response.text)

    term_to_gene = []
    if lib_name not in results:
        raise AssertionError("Something happened when calling enrichR.")

    for term, genes_dict in results[lib_name]['terms'].items():
        genes = sorted(set(i for i in genes_dict))
        term_to_gene.append(
            dict(term=term.lower(), gene_list=genes, n_genes=len(genes))
        )

    return term_to_gene


def run_enrichment_for_project(exp_data, project_name, databases=None,
                               output_path=None):
    """

    Parameters
    ----------
    exp_data : magine.data.experimental_data.ExprerimentalData
    project_name : str
    databases : list
    output_path : str
        Location to save all individual enrichment output files created.

    """

    if databases is None:
        databases = standard_dbs

    logger.info("Running enrichment on project")
    logger.info("Running {} databases".format(len(databases)))

    e = Enrichr(verbose=True)
    all_df = []
    if output_path is None:
        _dir = os.path.join(os.getcwd(), 'enrichment_output')
    else:
        _dir = output_path

    if not os.path.exists(_dir):
        logger.info("Creating output directory: {}".format(_dir))
        os.mkdir(_dir)

    def _run_new(samples, timepoints, category):
        logger.info("Running {}".format(category))
        for genes, sample_id in zip(samples, timepoints):
            if not len(genes):
                continue
            logger.info('\t time point = {}'.format(sample_id))
            current = "{}_{}_{}".format(category, sample_id, project_name)
            name = os.path.join(_dir, current + '.csv.gz')
            try:
                df = pd.read_csv(name, index_col=None, encoding='utf-8')
            except:
                df = e.run(genes, databases)
                df['sample_id'] = sample_id
                df['category'] = category
                df.to_csv(name, index=False, encoding='utf-8',
                          compression='gzip')
            df['sample_id'] = sample_id
            df['category'] = category
            all_df.append(df)

    #  run all protein labeled species grouped by time point
    #  ( label-free, ph-silac, etc are all combined)
    if len(exp_data.proteins.sample_ids) != 0:
        sample = exp_data.proteins.sig
        _run_new(sample.by_sample, sample.sample_ids, 'proteomics_both')
        _run_new(sample.up_by_sample, sample.sample_ids, 'proteomics_up')
        _run_new(sample.down_by_sample, sample.sample_ids, 'proteomics_down')

    #  run all RNA labeled species grouped by time point
    if len(exp_data.rna.sample_ids) != 0:
        sample = exp_data.rna.sig
        _run_new(sample.by_sample, sample.sample_ids, 'rna_both')
        _run_new(sample.down_by_sample, sample.sample_ids, 'rna_down')
        _run_new(sample.up_by_sample, sample.sample_ids, 'rna_up')

    #  run each experimental 'source' by time point
    #  ( label-free, ph-silac, etc are all separate)
    for source in exp_data.exp_methods:
        df = exp_data[source].sig
        col_options = df['species_type'].unique()
        # make sure there is only a single species type and it is protein
        # this makes sure that the RNA seq doesnt get ran twice
        if len(col_options) != 1:
            continue
        if col_options[0] == 'protein':
            _run_new(df.by_sample, df.sample_ids, '{}_both'.format(source))
            _run_new(df.up_by_sample, df.sample_ids, '{}_up'.format(source))
            _run_new(df.down_by_sample, df.sample_ids,
                     '{}_down'.format(source))

    # merge all outputs
    final_df = pd.concat(all_df, ignore_index=True)

    final_df = final_df[
        ['term_name', 'rank', 'combined_score', 'adj_p_value', 'genes',
         'n_genes', 'sample_id', 'category', 'db']
    ]

    # remove rows without a term name
    final_df = final_df[~final_df['term_name'].isnull()].copy()
    final_df = final_df[~final_df['adj_p_value'].isnull()].copy()

    # Adds significant column
    final_df.loc[(final_df['adj_p_value'] <= 0.05) &
                 (final_df['combined_score'] > 0.0), 'significant'] = True

    logger.info("Saving output: {}_enrichment.csv.gz".format(project_name))
    final_df.to_csv('{}_enrichment.csv.gz'.format(project_name),
                    encoding='utf-8', compression='gzip', index=False)



from .download_libraries import HMDB, download_hgnc, download_ncbi, \
    download_uniprot

if __name__ == '__main__':
    download_hgnc()
    download_ncbi()
    download_uniprot()
    HMDB().download_db()



import os
from collections import defaultdict

import pandas as pd
import requests
from magine.enrichment.ontology_analysis import MagineGO

from magine.data.storage import id_mapping_dir
from magine.data.storage import network_data_dir

try:
    import cPickle as pickle
except:
    import pickle


def download_and_process_go(species='hsa'):
    print("Creating GO files")
    from goatools import obo_parser
    obo_file = os.path.join(id_mapping_dir, 'go.obo')
    if not os.path.exists(obo_file):
        download_current_go()
    go = obo_parser.GODag(obo_file)
    gene_to_go, go_to_gene, goid_to_name = download_ncbi_gene_file()

    go_aspect = dict()
    go_depth = dict()

    dirname = network_data_dir

    go_to_gene_name = os.path.join(dirname,
                                   '{}_goids_to_genes.p'.format(species))
    go_to_go_name = os.path.join(dirname,
                                 '{}_goids_to_goname.p'.format(species))
    gene_to_go_name = os.path.join(dirname, '{}_gene_to_go.p'.format(species))
    go_depth_name = os.path.join(dirname, '{}_godepth.p'.format(species))
    go_aspect_name = os.path.join(dirname, '{}_go_aspect.p'.format(species))
    for i in go_to_gene.keys():
        go_depth[i] = go[i].depth
        go_aspect[i] = go[i].namespace

    pickle.dump(go_to_gene, open(go_to_gene_name, 'wb'))
    pickle.dump(goid_to_name, open(go_to_go_name, 'wb'))
    pickle.dump(go_depth, open(go_depth_name, 'wb'))
    pickle.dump(go_aspect, open(go_aspect_name, 'wb'))

    for i in go_to_gene:
        term = i
        genes = go_to_gene[i]
        for g in genes:
            if g in gene_to_go:
                gene_to_go[g].add(term)
            else:
                gene_to_go[g] = set()
                gene_to_go[g].add(term)
    pickle.dump(gene_to_go, open(gene_to_go_name, 'wb'))
    print("Done creating GO files")


def download_ncbi_gene_file(tax_ids=None):
    """ Downloads gene2go associations files from ncbi

    Parameters
    ----------
    tax_ids : list
        list of tax ids to consider, default [9606], human
    force_dnld

    Returns
    -------

    """

    from magine.mappings.gene_mapper import GeneMapper
    gm = GeneMapper()

    gene2go_url = "ftp://ftp.ncbi.nlm.nih.gov/gene/DATA/gene2go.gz"
    print("Downloading gene2go assocations file")
    r = pd.read_table(gene2go_url, compression='gzip')

    id2gos = defaultdict(set)
    go2term = dict()
    go2genes = defaultdict(set)

    if tax_ids is None:  # Default taxid is Human
        tax_ids = {9606}
    for line in r.values:
        t_id, gene_id, go_id, evidence, qual, go_term = line[:6]

        t_id = int(t_id)
        if t_id in tax_ids and qual != 'NOT' and evidence != 'ND':
            gene_id = int(gene_id)
            symbol = gm.ncbi_to_symbol[gene_id]
            if len(symbol) == 1:
                symbol = symbol[0]
            else:
                print(symbol)
            go2genes[go_id].add(symbol)
            id2gos[symbol].add(go_id)
            go2term[go_id] = go_term

    return id2gos, go2genes, go2term


def download_current_go(redownload=True):
    go_url = 'http://purl.obolibrary.org/obo/go.obo'
    target_file = 'go.obo'
    out_path = os.path.join(id_mapping_dir, target_file)
    if not os.path.exists(out_path) or redownload:
        print("GO exists, default behavior is to re-download.")

        r = requests.get(go_url, stream=True)
        response = requests.head(go_url)
        print(response.headers)
        # file_size = int(response.headers['content-length'])
        print("Downloading ontology file")
        file_size_dl = 0
        block_sz = 1024
        # block_sz = 8024

        with open(out_path, 'wb') as f:
            for chunk in r.iter_content(chunk_size=block_sz):
                file_size_dl += len(chunk)

                if chunk:  # filter out keep-alive new chunks
                    f.write(chunk)
        print("Downloaded {} and stored {}".format(go_url, out_path))


if __name__ == '__main__':
    # create_annotations()
    download_and_process_go()
    go = MagineGO('hsa')
    terms = go.calculate_enrichment(['BAX'],
                                    reference=['BAX', "LL", 'BCL2', 'CASP1'])
    #    'GO:0001569', (['BAX'], 0.78977569118414181, 1))
    for t in terms:
        print(t, terms[t])



from locale import getpreferredencoding

import matplotlib.pyplot as plt
import networkx as nx
import numpy as np
import pydot
from networkx.drawing.layout import rescale_layout

from magine.networks.exporters import nx_to_dot

try:
    basestring
except NameError:
    basestring = str
    unicode = str


def pydot_layout(G, prog='neato'):
    """Create node positions using :mod:`pydot` and Graphviz.

    Parameters
    --------
    G : Graph
        NetworkX graph to be laid out.
    prog : optional[str]
        Basename of the GraphViz command with which to layout this graph.
        Defaults to `neato`, the default GraphViz command for undirected graphs.

    Returns
    --------
    dict
        Dictionary of positions keyed by node.

    """
    tmp_g = G.copy()
    tmp_g.graph.update({'K': '1'})
    tmp_g.graph.update({'repulsiveforce ': '1'})
    tmp_g.graph.update({'overlap ': 'false'})
    tmp_g.graph.update({'splines ': 'true'})
    dot_fmt = nx_to_dot(tmp_g).create(prog=prog, format='dot', encoding='utf8')
    dot_fmt = unicode(dot_fmt, encoding=getpreferredencoding())
    net = pydot.graph_from_dot_data(dot_fmt)[0]
    node_pos = {}

    for n in tmp_g.nodes():
        pydot_node = pydot.Node(n).get_name()
        node = net.get_node(pydot_node)

        if isinstance(node, list):
            node = node[0]
        pos = node.get_pos()[1:-1]  # strip leading and trailing double quotes
        if pos is not None:
            xx, yy = pos.split(",")
            node_pos[n] = (float(xx), float(yy))

    return node_pos


def draw_mpl(network, layout='dot', scale=1, node_size=500, font_size=16):
    """ Draw network using networkx and matplotlib

    Parameters
    ----------
    network : nx.DiGraph
    layout : str

    """
    for (n, d) in network.nodes(data=True):
        if 'keggName' in d:
            del d["keggName"]
    mpl_layouts = ['circular_layout', 'random_layout', 'shell_layout',
                   'spring_layout', 'spectral_layout', 'kamada_kawai_layout',
                   'fruchterman_reingold_layout']
    dot_layouts = ['dot', 'neato', 'fdp', 'twopi', 'circo', 'sfdp']
    if layout in dot_layouts:
        pos = pydot_layout(network, prog=layout, )
    else:
        if layout not in mpl_layouts:
            raise AssertionError("Please provide valid layout option:{}"
                                 "".format(mpl_layouts + dot_layouts))
        if layout == 'spring_layout':
            pos = nx.drawing.layout.spring_layout(network)
        elif layout == 'random_layout':
            pos = nx.drawing.layout.random_layout(network)
        elif layout == 'shell_layout':
            pos = nx.drawing.layout.shell_layout(network)
        elif layout == 'spectral_layout':
            pos = nx.drawing.layout.spectral_layout(network)
        elif layout == 'fruchterman_reingold_layout':
            pos = nx.drawing.layout.fruchterman_reingold_layout(network)
        elif layout == 'circular_layout':
            pos = nx.drawing.layout.circular_layout(network)
        elif layout == 'kamada_kawai_layout':
            pos = nx.drawing.layout.kamada_kawai_layout(network)

    # some layout algorithms (graphviz ones) can provide large position values
    # normalize the positions to be from 0 to 1
    # this allows us to add space between pode and label
    positions = np.array([p for _, p in pos.items()])

    positions = rescale_layout(positions, scale=1)
    new_pos, label_pos = {}, {}
    for num, n in enumerate(pos):
        x, y = positions[num]
        new_pos[n] = [x, y]
        # raise text positions
        label_pos[n] = [x, y + .05]
    try:
        node_colors = [n['color'] for _, n in network.nodes(data=True)]
    except:
        node_colors = ['red' for _ in network.nodes()]
    fig = plt.figure(figsize=(4 * scale, 4 * scale))
    ax = fig.add_subplot()
    nx.draw_networkx(network, pos=new_pos,
                     with_labels=False,
                     node_color=node_colors,
                     node_size=node_size,
                     alpha=.4, ax=ax
                     )
    nx.draw_networkx_labels(network, label_pos, font_size=font_size, ax=ax)
    # plt.axis('equal')
    plt.axis('off')
    # plt.xticks([])
    # plt.yticks([])
    # plt.tight_layout()
    return fig


if __name__ == '__main__':
    x = nx.DiGraph()
    x.add_edge('A', 'B')
    x.add_edge('A', 'C')
    x.add_edge('C', 'D')
    x.add_edge('C', 'E')
    x.add_edge('E', 'A')
    # x = pydot_layout(x)
    draw_mpl(x, 'circo', scale=1)
    draw_mpl(x, 'circo', scale=2)
    plt.show()



import numpy as np


def log2_normalize_df(df, column, new_col=None):
    """

    Parameters
    ----------
    df : pandas.DataFrame
        dataframe of fold changes
    column : str
        column that contains fold change values
    new_col : str, None
        Name of new column to place fold_changes values. if not provided it
        will overwrite column values with new values
    Returns
    -------

    """
    tmp_df = df.copy()
    greater_than = tmp_df[column] > 0
    less_than = tmp_df[column] < 0
    if new_col is not None:
        if not isinstance(new_col, str):
            raise AssertionError()
        tmp_df.loc[greater_than, new_col] = np.log2(
            tmp_df[greater_than][column])
        tmp_df.loc[less_than, new_col] = -np.log2(-tmp_df[less_than][column])
    else:
        tmp_df.loc[greater_than, column] = np.log2(
            tmp_df[greater_than][column])
        tmp_df.loc[less_than, column] = -np.log2(-tmp_df[less_than][column])
    return tmp_df



import os
import shutil
import tempfile

import matplotlib.pyplot as plt
import pandas as pd
from nose.tools import ok_

from magine.data.experimental_data import ExperimentalData, load_data


class TestExpData(object):
    def setUp(self):
        self._dir = os.path.join(os.path.dirname(__file__), 'Data')
        self.exp_data = load_data(
            os.path.join(self._dir, 'example_apoptosis.csv')
        )
        self.out_dir = tempfile.mkdtemp()

    def tearDown(self):
        self.exp_data = None
        shutil.rmtree(self.out_dir)

    def test_load_from_df(self):
        df = pd.read_csv(os.path.join(self._dir, 'example_apoptosis.csv'))
        ExperimentalData(df)

    def test_protein(self):
        ok_(self.exp_data.proteins.id_list == {'AHR', 'PARP4', 'ADORA1',
                                               'BAX', 'TP53', 'PARP1',
                                               'ADRA1A', 'ADORA2A', 'CASP3',
                                               'AGTR2', 'BID'})

        ok_(self.exp_data.proteins.sig.id_list == {'ADRA1A', 'BAX', 'TP53',
                                                   'AGTR2', 'PARP4', 'BID',
                                                   'PARP1', 'CASP3'})

        ok_(self.exp_data.proteins.up.id_list == {'PARP4', 'BAX', 'PARP1',
                                                  'BID', 'CASP3', 'TP53',
                                                  'ADRA1A'})

        ok_(self.exp_data.proteins.down.id_list == {'AGTR2', 'BAX'})

    def test_gene(self):
        ok_(self.exp_data.genes.id_list == {'ADORA2A', 'CASP3', 'PARP4',
                                            'PARP1', 'ADORA1', 'BAX',
                                            'ADRA1A', 'AIF1', 'AKT2',
                                            'AGTR2', 'AHR', 'TP53', 'BID',
                                            'AKT1'})

        ok_(self.exp_data.genes.sig.id_list == {'CASP3', 'PARP4', 'PARP1',
                                                'BAX', 'AIF1', 'ADRA1A',
                                                'AGTR2', 'TP53', 'BID', 'AKT1'}
            )

    def test_rna(self):
        ok_(self.exp_data.rna.id_list == {'AIF1', 'AKT1', 'AKT2'})
        ok_(self.exp_data.rna.sig.id_list == {'AIF1', 'AKT1'})

    def test_compounds(self):
        compounds = {'HMDB0009901', 'HMDB0000001'}
        ok_(self.exp_data.compounds.sig.id_list == compounds)
        ok_(self.exp_data.compounds.id_list == compounds)

    def test_species(self):
        ok_(self.exp_data.species.id_list == {'BID', 'TP53', 'PARP4',
                                              'HMDB0009901', 'BAX', 'AKT1',
                                              'AKT2', 'PARP1', 'AGTR2',
                                              'AHR', 'AIF1', 'ADORA1',
                                              'CASP3', 'HMDB0000001',
                                              'ADORA2A', 'ADRA1A'})

        ok_(self.exp_data.species.sig.id_list == {'BID', 'PARP4', 'BAX',
                                                  'AKT1', 'PARP1', 'AGTR2',
                                                  'AIF1', 'CASP3', 'TP53',
                                                  'HMDB0000001', 'ADRA1A',
                                                  'HMDB0009901'})

    def test_plot_list(self):
        self.exp_data.rna.plot_species(self.exp_data.rna.sig.id_list,
                                       save_name='del_test',
                                       out_dir=self.out_dir)
        plt.close()

    def test_html_output(self):
        self.exp_data.proteins.plot_all(html_file_name='del',
                                        out_dir=self.out_dir)
        plt.close()

    def test_heatmap(self):
        self.exp_data.proteins.heatmap()
        plt.close()
        self.exp_data.species.heatmap({'BID', 'PARP4', 'BAX', 'AKT1', 'PARP1',
                                       'AGTR2', 'AIF1', 'CASP3', 'TP53',
                                       'HMDB0000001', 'ADRA1A', 'HMDB0009901'})
        plt.close()
        self.exp_data.species.heatmap({'BID', 'PARP4', 'BAX', 'AKT1', 'PARP1',
                                       'AGTR2', 'AIF1', 'CASP3', 'TP53',
                                       'HMDB0000001', 'ADRA1A', 'HMDB0009901'},
                                      cluster_col=True, cluster_row=True)
        plt.close()
        self.exp_data.species.heatmap({'BID', 'PARP4', 'BAX', 'AKT1', 'PARP1',
                                       'AGTR2', 'AIF1', 'CASP3', 'TP53',
                                       'HMDB0000001', 'ADRA1A', 'HMDB0009901'},
                                      columns=['sample_id', 'source'])

    def test_volcano(self):
        self.exp_data.volcano_analysis(out_dir=self.out_dir)
        plt.close()

    def test_time_series_volcano(self):
        self.exp_data.label_free.volcano_by_sample('test_label_free',
                                                   out_dir=self.out_dir,
                                                   sig_column=True)
        plt.close()

    def test_plot_all_metabolites(self):
        self.exp_data.compounds.plot_all('metab', out_dir=self.out_dir,
                                         plot_type='matplotlib')
        plt.close()

    def test_list_metabolites(self):
        l = ['HMDB0000001', 'HMDB0009901']
        self.exp_data.compounds.plot_species(l, save_name='metab',
                                             out_dir=self.out_dir,
                                             plot_type='plotly')
        plt.close()

    def test_histogram(self):
        self.exp_data.label_free.plot_histogram(
            save_name='lf_hist', y_range=[0, 100], out_dir=self.out_dir
        )
        plt.close()

    def test_table(self):
        self.exp_data.create_summary_table()
        self.exp_data.create_summary_table(write_latex=True, save_name='latex')
        self.exp_data.create_summary_table(sig=True)
        self.exp_data.create_summary_table(sig=True, index='label')

    def test_log2(self):
        x = self.exp_data.rna.log2_normalize_df('fold_change')
        ok_(x.to_dict() ==
            {'sample_id': {8: 'Time_3', 9: 'Time_3', 10: 'Time_3'},
             'source': {8: 'rna_seq', 9: 'rna_seq', 10: 'rna_seq'},
             'significant': {8: True, 9: True, 10: False},
             'fold_change': {8: -0.5849625007211562,
                             9: -1.8073549220576042,
                             10: -0.2630344058337938},
             'p_value': {8: 0.01, 9: 0.06, 10: 0.22},
             'identifier': {8: 'AIF1', 9: 'AKT1', 10: 'AKT2'},
             'species_type': {8: 'protein', 9: 'protein', 10: 'protein'},
             'label': {8: 'AIF1', 9: 'AKT1', 10: 'AKT2'}})

    def test_pivotor(self):
        x = self.exp_data.rna.pivoter(True, columns='sample_id',
                                      values='fold_change', fill_value='-',
                                      min_sig=1)
        ok_(x.to_dict('index') == {'AKT1': {'Time_3': -1.8073549220576042},
                                   'AIF1': {'Time_3': -0.5849625007211562}})

    def test_subset(self):
        species = ['AKT1', 'AIF1']
        x = self.exp_data.subset(species, index='identifier')
        ok_(x.shape == (2, 8))



import os
from shutil import copytree, rmtree

from appdirs import user_data_dir


def copy_sample_databases(force=False):
    dir_name = user_data_dir('magine')

    network_data_dir = os.path.join(dir_name, 'network_data')
    id_mapping_dir = os.path.join(dir_name, 'id_data')

    # Check if database directories already exist
    if os.path.exists(id_mapping_dir) and os.path.exists(network_data_dir) and not force:
        print('Database directories already exist. Override using force=True argument. Skipping...')
        return

    # check or create the main data storage directory
    if not os.path.exists(dir_name):
        os.makedirs(dir_name)

    # check or create the ID mapping data storage directory
    if os.path.exists(id_mapping_dir):
        rmtree(id_mapping_dir)

    # check or create the network data storage directory
    if os.path.exists(network_data_dir):
        rmtree(network_data_dir)

    _sample_dir = os.path.join(os.path.dirname(__file__), '..',
                               '_sample_databases')
    id_dir = os.path.join(_sample_dir, 'id_data')
    network_dir = os.path.join(_sample_dir, 'network_data')

    copytree(id_dir, id_mapping_dir)
    copytree(network_dir, network_data_dir)


if __name__ == '__main__':
    copy_sample_databases()



import os

import networkx as nx

from magine.enrichment.enrichr import Enrichr
from magine.networks.annotated_set import create_subnetwork


def test_ont_grouping():
    e = Enrichr()

    list_1 = ['CASP3', 'CASP6', 'FAS', 'FADD', 'CASP8', 'CFLAR', 'BFAR', 'BAD',
              'BID', 'PMAIP1', 'MCL1', 'BCL2', 'BCL2L1', 'BAX', 'BAK1',
              'DIABLO', 'CYCS', 'PARP1', 'APAF1', 'XIAP']

    list_2 = ['DIABLO', 'CYCS', 'PARP1', 'APAF1', 'XIAP']

    df = e.run_samples([list_1, list_2], 'GO_Biological_Process_2017')
    df = df.sig.copy()
    _path = os.path.join(os.path.dirname(__file__), 'Network_files',
                         'sample_network.gml')
    network = nx.read_gml(_path)
    create_subnetwork(df, network, use_threshold=True, save_name='test')
    create_subnetwork(df, network, use_threshold=True, use_fdr=True)
    create_subnetwork(df, network, use_threshold=True, use_fdr=True,
                      out_dir='del')



import os

import jinja2

env = jinja2.Environment(
    loader=jinja2.FileSystemLoader(
        searchpath=os.path.join(os.path.dirname(__file__), 'templates')
    ),
    autoescape=False
)

workflow_template = env.get_template('workflow_template.html')
plotly_template = env.get_template('plotly_template.html')
single_template = env.get_template('single_table_view.html')
filter_template = env.get_template('filter_table.html')

range_ = {'ref', 'depth', 'enrichment_score', 'rank', 'fold_change',
          'p_value', 'adj_p_value', 'combined_score', 'n_genes',
          'z_score', 'pvalue', 'treated_control_fold_change',
          'p_value_group_1_and_group_2'}

chosen_ = {
    'GO_name', 'slim', 'aspect', 'time', 'data_type', 'identifier', 'label',
    'term_name', 'term_id', 'category', 'db',  # enrichr
    'significant_flag', 'project_name', 'compound', 'GO_id',
    'sample_id', 'sample_index'
}
multi_choose = {'genes', 'source', 'significant'}

auto_complete_ = {'protein', 'gene', 'identifier', 'label', }


def _add_filter(column_num, f_type):
    _default = {'column_number': column_num}
    if f_type == 'range':
        _default.update({'filter_type': 'range_number'})
    elif f_type == 'select':
        _default.update({
            'select_type_options': {'width': "150px"},
            'filter_type': 'multi_select',
            'select_type': 'select2'
        })
    elif f_type == 'chosen':
        _default.update({'select_type': "multi_select",
                         'select_type_options': {'width': "150px"},
                         'text_data_delimiter': "chosen"},
                        )
    elif f_type == 'auto_complete':
        _default.update({'filter_type': "auto_complete",
                         'text_data_delimiter': ","})
    elif f_type == 'multichoose':
        _default.update({
            'filter_type': "multi_select",
            'select_type': "select2",
            'select_type_options': {'width': "150px"},
            'text_data_delimiter': ","},
        )
    return _default


def create_yadf_filters(table):
    _format_dict = []
    for n, i in enumerate(table.columns):
        if isinstance(i, (tuple, list)):
            i = i[0]
        if i in range_:
            _format_dict.append(_add_filter(n, 'range'))
        elif i in chosen_:
            _format_dict.append(_add_filter(n, 'select'))
        elif i in auto_complete_:
            _format_dict.append(_add_filter(n, 'complete'))
        elif i in multi_choose:
            _format_dict.append(_add_filter(n, 'multichoose'))
        else:
            print("'{}' not found in yadf".format(i))
    return _format_dict


def write_filter_table(table, save_name):
    """

    Parameters
    ----------
    table : pandas.DataFrame
    save_name : str

    Returns
    -------

    """

    # formats output to less precision and ints rather than floats
    tmp_table = _format_simple_table(table)

    data = tmp_table.to_dict('split')
    data['filters'] = create_yadf_filters(table)
    html_out = filter_template.render({"data": data})
    with open('{}.html'.format(save_name), 'w') as f:
        f.write(html_out)


def _format_simple_table(data):
    """
    formats precession of data for outputs

    Parameters
    ----------
    data : pandas.DataFrame

    Returns
    -------
    pandas.DataFrame, dict
    """
    tmp_table = data.copy()
    pvalue_float_type = ['pvalue', 'p_value', 'p_value_group_1_and_group_2',
                         'adj_p_value']

    float_type = ['z_score', 'combined_score', 'enrichment_score',
                  'treated_control_fold_change']

    int_type = ['n_genes', 'rank']
    for i in data.columns:
        if isinstance(i, (tuple, list)):
            t_i = i[0]
        else:
            t_i = i
        if t_i in float_type:
            tmp_table[i] = tmp_table[i].fillna(0)
            tmp_table[i] = tmp_table[i].astype(float)
            tmp_table[i] = tmp_table[i].round(2)
            tmp_table[i] = tmp_table[i].apply('{:.4g}'.format)
        elif t_i in pvalue_float_type:
            tmp_table[i] = tmp_table[i].fillna(1)
            tmp_table[i] = tmp_table[i].apply('{:.2g}'.format)
        elif t_i in int_type:
            tmp_table[i] = tmp_table[i].fillna(0)
            tmp_table[i] = tmp_table[i].astype(int)
            tmp_table[i] = tmp_table[i].apply('{:,d}'.format)
    return tmp_table


def format_ploty(text, save_name):
    """
    
    Parameters
    ----------
    text : str
        html code to embed in file
    save_name : str
        html output filename

    Returns
    -------

    """

    html_out = plotly_template.render({"plotly_code": text})
    with open('{}'.format(save_name), 'w') as f:
        f.write(html_out)



import os

import matplotlib.pyplot as plt
import numpy as np

from magine.data.tools import log2_normalize_df

fold_change = 'fold_change'
flag = 'significant'
exp_method = 'source'
p_val = 'p_value'
rna = 'rna_seq'
gene = 'gene'
protein = 'protein'
metabolites = 'metabolites'
species_type = 'species_type'
sample_id = 'sample_id'


def create_mask(data, use_sig=True, p_value=0.1, fold_change_cutoff=1.5):
    """ Creates a mask for volcano plots.


    # Visual example of volcano plot
    # section 0 are significant criteria

    #    0    #    1   #      0     #
    #         #        #            #
    #################################
    #         #        #            #
    #    2    #    2   #      2     #
    #         #        #            #
    #################################


    Parameters
    ----------
    data : pd.DataFrame
    use_sig : bool
    p_value : float
        p_value threshold
    fold_change_cutoff : float
        fold change threshold

    Returns
    -------

    """
    # copy of data
    tmp = data.loc[:, (p_val, fold_change, flag)].copy()

    # convert to log10 scale
    tmp[p_val] = np.log10(data[p_val]) * -1

    # convert to log2 space
    tmp = log2_normalize_df(tmp, column=fold_change)

    if use_sig:
        sec_0 = tmp[tmp[flag]]
        sec_2 = tmp[~tmp[flag]]
        sec_1 = None
    else:
        fc = np.log2(fold_change_cutoff)
        p_value = -1 * np.log10(p_value)
        criteria_1 = tmp[p_val] >= p_value
        sec_0 = tmp[criteria_1 & (np.abs(tmp[fold_change]) >= fc)]
        sec_1 = tmp[criteria_1 & (np.abs(tmp[fold_change]) < fc)]
        sec_2 = tmp[(tmp[p_val] < p_value)]
    return sec_0, sec_1, sec_2


def add_volcano_plot(fig_axis, section_0, section_1, section_2):
    """ Adds a volcano plot to a fig axis

    Parameters
    ----------
    fig_axis : plt.Figure.axes
    section_0 : pd.DataFrame
    section_1 : pd.DataFrame
    section_2 : pd.DataFrame

    Returns
    -------

    """
    fig_axis.scatter(section_0[fold_change], section_0[p_val], marker='.',
                     color='blue')
    if section_1 is not None:
        fig_axis.scatter(section_1[fold_change], section_1[p_val],
                         marker='.', color='red')
    fig_axis.scatter(section_2[fold_change], section_2[p_val], s=1,
                     marker=',', color='gray')
    fig_axis.set_ylabel('-log$_{10}$ p-value', fontsize=16)
    fig_axis.set_xlabel('log$_2$ Fold Change', fontsize=16)


def volcano_plot(data, save_name=None, out_dir=None, sig_column=False,
                 p_value=0.1, fold_change_cutoff=1.5, x_range=None,
                 y_range=None):
    """ Create a volcano plot of data

    Creates a volcano plot of data type provided

    Parameters
    ----------
    data : pandas.DataFrame
        data to create volcano plots from
    save_name: str
        name to save figure
    out_dir: str, directory
        Location to save figure
    sig_column: bool, optional
        If to use significant flags of data
    p_value: float, optional
        Criteria for significant
    fold_change_cutoff: float, optional
        Criteria for significant
    y_range: array_like
        upper and lower bounds of plot in y direction
    x_range: array_like
        upper and lower bounds of plot in x direction

    Returns
    -------


    """
    cp_df = data.copy()
    cp_df.dropna(subset=[p_val], inplace=True)
    cp_df = cp_df[np.isfinite(data[fold_change])]
    sec_0, sec_1, sec_2 = create_mask(cp_df, sig_column, p_value,
                                      fold_change_cutoff)
    fig = plt.figure()
    ax = fig.add_subplot(111)
    add_volcano_plot(ax, sec_0, sec_1, sec_2)

    if not sig_column:
        fc = np.log2(fold_change_cutoff)
        log_p_val = -1 * np.log10(p_value)
        ax.axvline(x=fc, linestyle='--')
        ax.axvline(x=-1 * fc, linestyle='--')
        ax.axhline(y=log_p_val, linestyle='--')
    if y_range is not None:
        ax.set_ylim(y_range[0], y_range[1])
    if x_range is not None:
        ax.set_xlim(x_range[0], x_range[1])
    fig.tight_layout()
    if save_name is not None:
        save_plot(fig, save_name=save_name, out_dir=out_dir)
    return fig


def save_plot(fig, save_name, out_dir=None, image_type='png'):
    """ Saves fig
    
    Parameters
    ----------
    fig : plt.Figure
        Figure to be saved
    save_name : str
        output file name
    out_dir : str, optional
        output path
    image_type : str, optional
        output type of file, {"png", "pdf", etc..}

    Returns
    -------

    """
    fig.tight_layout()
    save_name = '{}.{}'.format(save_name, image_type)
    if out_dir is not None:
        save_name = os.path.join(out_dir, save_name)
    fig.savefig(save_name, bbox_inches='tight')



from functools import partial
import itertools

import networkx as nx

from magine.networks.visualization.graphviz import draw_graphviz, \
    paint_network_overtime_up_down, paint_network_overtime


class Subgraph(object):
    def __init__(self, network, exp_data=None, pool=None):
        """
        Generates network subgraphs

        Parameters
        ----------
        network : networkx.DiGraph
        exp_data : magine.data.datatypes.ExperimentalData
        
        """
        self.network = network
        self.nodes = set(self.network.nodes())
        self.exp_data = exp_data

        if pool is None:
            self.map = map
        else:
            self.map = pool.map

    def paths_between_pair(self, node_1, node_2, bidirectional=False,
                           single_path=False, draw=False, image_format='png'):
        """
        Generates a graph based on all shortest paths between two species.


        Parameters
        ----------
        node_1 : str
            name of first species
        node_2 : str
            name of second species
        bidirectional : bool
            If you want to search bidirectionally
        single_path : bool
            If you only want a single shortest path
        draw : bool
            create an image of returned network
        image_format : str, optional
            If draw=True you can pass an image format. (pdf, png, svg). 
            default=png

        Returns
        -------
        graph : networkx.DiGraph


        Examples
        --------
        >>> from networkx import DiGraph
        >>> from magine.networks.subgraphs import Subgraph
        >>> g = DiGraph()
        >>> g.add_edges_from([('a','b'),('b','c'), ('c', 'd'),  ('e', 'd'), ('d', 'a')])
        >>> net_sub = Subgraph(g)
        >>> path_a_d = net_sub.paths_between_pair('a','d', False)
        >>> sorted(path_a_d.edges)
        [('a', 'b'), ('b', 'c'), ('c', 'd')]
        >>> path_a_d = net_sub.paths_between_pair('a','d', True)
        >>> sorted(path_a_d.edges)
        [('a', 'b'), ('b', 'c'), ('c', 'd'), ('d', 'a')]
        """

        for i in (node_1, node_2):
            if i not in self.nodes:
                print("{} is not in network!")
                return

        paths = [_nx_find_path(self.network, node_1, node_2, single_path)]
        if bidirectional:
            paths += [_nx_find_path(self.network, node_2, node_1, single_path)]
        graph = self._process(paths, None, None, False)
        if draw:
            save_name = "%s_and_%s" % (node_1, node_2)
            self._save_or_draw(graph, save_name, draw, image_format)

        return graph

    def paths_between_two_lists(self, list_1, list_2, single_path=False,
                                max_length=None, include_only=None,
                                reverse=False, add_interconnecting_edges=False,
                                draw=False, save_name=None, image_format='png'
                                ):
        """
        Generates a graph based on all shortest paths between two species.


        Parameters
        ----------
        list_1 : list
            Node names
        list_2 : list
            Node names
        single_path : bool
            If you only want a single shortest path.
        max_length : int
            Maximum distance between any two species.
        include_only : list
            Species required to be in paths/
        reverse : bool
            Flag to check list_2 to list_1. Default will only look for list_1
            to list_2.
        add_interconnecting_edges : bool
            Add edges between species even if not between list_1 and list_2
            nodes.
        save_name : str
            Save of figure/network
        draw : bool
            create an image of returned network
        image_format : str, optional
            If draw=True you can pass an image format. (pdf, png, svg).
            default=png

        Returns
        -------
        graph : networkx.DiGraph

        Examples
        --------
        >>> from networkx import DiGraph
        >>> from magine.networks.subgraphs import Subgraph
        >>> g = DiGraph()
        >>> g.add_path(['a', 'b', 'c', 'd'])
        >>> g.add_path(['g', 'h', 'c', 'i'])
        >>> net_sub = Subgraph(g)
        >>> path_a_d = net_sub.paths_between_two_lists(['a','g'], ['c', 'i'], max_length=3)
        >>> sorted(path_a_d.edges)
        [('a', 'b'), ('b', 'c'), ('g', 'h'), ('h', 'c')]

        """

        tmp_list1 = list(self._check_node(list_1))
        tmp_list2 = list(self._check_node(list_2))
        paths = [_nx_find_path(self.network, i, j, single_path)
                 for i, j in itertools.product(tmp_list1, tmp_list2)]
        if reverse:
            paths += [_nx_find_path(self.network, i, j, single_path)
                      for i, j in itertools.product(tmp_list2, tmp_list1)]
        if include_only is not None:
            include_copy = list(include_only) + list(list_1) + list(list_2)
        else:
            include_copy = None
        graph = self._process(paths, max_length, include_copy,
                              add_interconnecting_edges)

        if save_name is not None:
            self._save_or_draw(graph, save_name, draw, image_format)
        return graph

    def _process(self, paths, max_length, include_only,
                 add_interconnecting_edges):
        if max_length is not None:
            paths = self._max_distance(paths, max_length)
        _new_paths = []
        for path in paths:
            for p in path:
                _new_paths += [(p[i], p[i + 1]) for i in range(len(p) - 1)]
        _new_paths = set(_new_paths)
        graph = self.network.edge_subgraph(_new_paths).copy()
        if add_interconnecting_edges:
            graph = self.network.subgraph(graph.nodes).copy()
        if include_only is not None:
            graph = self._include_only(graph, include_only)
        return graph

    def paths_between_list(self, species_list, single_path=False,
                           max_length=None, add_interconnecting_edges=False,
                           include_only=None, pool=None,
                           save_name=None, draw=False, image_format='png',
                           ):
        """
        Returns graph containing all shortest paths between list.

        Parameters
        ----------

        species_list : list_like
            list of species
        save_name : str
            name to save
        single_path : bool
            use single shortest path if True, else use all shortest paths
        draw : bool
            create a dot generated figure
        image_format : str
            dot acceptable output formats, (pdf, png, etc)
        pool : multiprocessing.Pool
            If it it provided, it uses its map function to run this function.
        max_length : int
            Max length for path between any 2 species
        include_only : list_like
            List of species that must be present
        Returns
        -------
        graph : networkx.DiGraph
            graph containing paths between species list provided


        Examples
        --------
        >>> from networkx import DiGraph
        >>> from magine.networks.subgraphs import Subgraph
        >>> g = DiGraph()
        >>> g.add_edges_from([('a','b'),('b','c'), ('c', 'd'), ('a', 'd'), ('e', 'd')])
        >>> g.add_path(['g', 'h', 'c', 'i', 'j', 'k'])
        >>> net_sub = Subgraph(g)
        >>> path_a_d = net_sub.paths_between_list(['a','c','d'])
        >>> sorted(path_a_d.edges)
        [('a', 'b'), ('a', 'd'), ('b', 'c'), ('c', 'd')]
        >>> path_a_f = net_sub.paths_between_list(['g', 'h', 'j'], max_length=4)
        >>> sorted(path_a_f.edges)
        [('c', 'i'), ('g', 'h'), ('h', 'c'), ('i', 'j')]
        """

        tmp_species_list = list(self._check_node(species_list))

        if pool is not None:
            _map = pool.map
        else:
            _map = map
        paths = _map(partial(_find_nx_path, network=self.network,
                             single_path=single_path),
                     itertools.combinations(tmp_species_list, 2))

        if include_only is not None:
            include_copy = list(include_only) + list(species_list)
        else:
            include_copy = None
        graph = self._process(paths, max_length, include_copy,
                              add_interconnecting_edges)

        if save_name is not None:
            self._save_or_draw(graph, save_name, draw, image_format)
        return graph

    def neighbors(self, node, upstream=True, downstream=True, max_dist=1,
                  include_only=None, start_network=None):
        """
        Create network containing provided node and its neighbors.

        Parameters
        ----------
        node : str
        upstream : bool
        downstream : bool
        max_dist : int
        include_only : list
        start_network : nx.DiGraph


        Returns
        -------
        nx.DiGraph
        """

        if not upstream and not downstream:
            print("Must provide up_stream=True or down_stream=True. "
                  "Returning None")
            return None

        if start_network is None:
            sg = nx.DiGraph()
        else:
            sg = start_network.copy()

        if node not in self.nodes:
            print("Node not in graph")
            return sg

        if max_dist > 3:
            print("Max distance is larger than 3. Warning! "
                  "Result will be largenetworks")

        def _add_node(n):
            sg.add_node(n, **self.network.node[n])

        def _add_edge(i, j):
            sg.add_edge(i, j, **self.network.edges[i, j])

        def _get_upstream(new_node):
            up_nodes = list(self.network.predecessors(new_node))
            for i in up_nodes:
                if i != new_node:
                    _add_node(i)
                    _add_edge(i, new_node)
            return up_nodes

        def _get_downstream(new_node):
            down_nodes = list(self.network.successors(new_node))
            for i in down_nodes:
                if i != new_node:
                    _add_node(i)
                    _add_edge(new_node, i)
            return down_nodes

        _add_node(node)

        up_layer = [node]
        down_layer = [node]
        for _ in range(0, max_dist):
            if upstream:
                up_layer = {i for n in up_layer for i in _get_upstream(n)}
            if downstream:
                down_layer = {i for n in down_layer for i in
                              _get_downstream(n)}
        if include_only is not None:
            include_copy = list(include_only)
            if node not in include_only:
                include_copy += [node]
            sg = self._include_only(sg, include_copy)

        return sg

    def expand_neighbors(self, network=None, nodes=None, upstream=False,
                         downstream=False, max_dist=1, include_only=None,
                         add_interconnecting_edges=False):
        """ Create/expand a network based on neighbors from a list of species

        Parameters
        ----------
        network : nx.DiGraph or None
            Starting network to expand nodes. If not provided, will use
            default network
        nodes : list_like
            List of nodes to expand
        upstream : bool
            Expand upstream nodes
        downstream : bool
            Expand downstream nodes
        max_dist :
            Max distance to explore
        include_only : list_like
            Limit network to only contain these species
        add_interconnecting_edges : bool
            Add edges connecting all nodes. Default if False, so only direct
            edges to neighbors will be added.

        Returns
        -------
        nx.DiGraph
        """

        if not upstream and not downstream:
            print("Must provide upstream=True or downstream=True. "
                  "Returning None")
            return None

        if network is None:
            new_g = nx.DiGraph()
            starting_nodes = set()
        else:
            new_g = network.copy()
            starting_nodes = set(new_g.nodes)

        if nodes is None:
            if network is None:
                raise AssertionError("Must provide network or list of nodes")
            nodes = set(new_g.nodes)
        elif isinstance(nodes, str):
            nodes = [nodes]
        elif not isinstance(nodes, (list, set)):
            print("Must provide node, list of nodes, or expand_all=True")
            return network

        for start in nodes:
            new_g = self.neighbors(
                start, start_network=new_g, max_dist=max_dist,
                upstream=upstream, downstream=downstream
            )

        if include_only is not None:
            include_only = set(include_only)
            include_only.update(set(nodes))
            include_only.update(starting_nodes)
            new_g = self._include_only(new_g, include_only)
        if add_interconnecting_edges:
            new_g = self.network.subgraph(new_g.nodes()).copy()
        return new_g

    def upstream_of_node(self, species_1, include_list=None,
                         save_name=None, draw=False):
        """ Generate network of all upstream species of provides species

        Parameters
        ----------
        species_1 : str
            species name
        save_name : str
            name to save gml file
        draw : bool
            create figure of graph
        include_list : list_like
            Species that must be in path in order to consider a path

        Returns
        -------
        nx.DiGraph


        Examples
        --------
        >>> from networkx import DiGraph
        >>> from magine.networks.subgraphs import Subgraph
        >>> g = DiGraph()
        >>> g.add_edges_from([('a','b'),('b','c'), ('c', 'd'), ('a', 'd'), \
        ('e', 'd')])
        >>> net_sub = Subgraph(g)
        >>> upstream_d = net_sub.upstream_of_node('d')
        >>> sorted(upstream_d.edges())
        [('a', 'd'), ('b', 'c'), ('c', 'd'), ('e', 'd')]
        >>> upstream_c = net_sub.upstream_of_node('c')
        >>> sorted(upstream_c.edges())
        [('a', 'b'), ('b', 'c')]

        """
        if include_list is None:
            include_list = set(self.nodes.copy())
        else:
            include_list = self._check_node(include_list)

        graph = self._list_paths_to_graph(
            [_nx_find_path(self.network, i, species_1, single_path=True)
             for i in self.nodes if i in include_list])
        if save_name is not None:
            self._save_or_draw(graph, save_name, draw)

        return graph

    def downstream_of_node(self, species_1, include_list=None,
                           save_name=None, draw=False):
        """ Generate network of all downstream species of provides species


        Parameters
        ----------
        species_1 : str
            species name
        save_name : str
            name to save gml file
        draw : bool
            create figure of graph
        include_list : list_like
            list of species that must be in path in order to consider a path
        Returns
        -------
        nx.DiGraph


        Examples
        --------
        >>> from networkx import DiGraph
        >>> from magine.networks.subgraphs import Subgraph
        >>> g = DiGraph()
        >>> g.add_edges_from([('a','b'),('b','c'), ('c', 'd'), ('a', 'd'),\
         ('e', 'd')])
        >>> net_sub = Subgraph(g)
        >>> downstream_d = net_sub.downstream_of_node('d')
        >>> sorted(downstream_d.edges)
        []
        >>> downstream_c = net_sub.downstream_of_node('c')
        >>> sorted(downstream_c.edges)
        [('c', 'd')]

        """
        if include_list is None:
            include_list = set(self.nodes.copy())
        else:
            include_list = self._check_node(include_list)

        graph = self._list_paths_to_graph(
            [_nx_find_path(self.network, species_1, i, single_path=True)
             for i in self.nodes if i in include_list])

        if save_name is not None:
            self._save_or_draw(graph, save_name, draw)

        return graph

    def measured_networks_over_time(self, graph, colors, prefix):
        """ Adds color to a network over time
        
        Parameters
        ----------
        graph : nx.DiGraph
        colors : list
            List of colors for time points
        prefix : str
            Prefix for image files
            

        Returns
        -------

        """
        measured_list = self.exp_data.species.sig.by_sample
        paint_network_overtime(graph, measured_list, colors, prefix,
                               self.exp_data.sample_ids)

    def measured_networks_over_time_up_down(self, graph, prefix,
                                            color_up='tomato',
                                            color_down='lightblue'):
        """

        Parameters
        ----------
        graph : nx.DiGraph

        prefix : str
            Prefix for image files
        color_up : str
        
        color_down : str


        Returns
        -------

        """
        labels = self.exp_data.sample_ids
        up_measured_list = self.exp_data.species.sig.up.by_sample
        down_measured_list = self.exp_data.species.sig.down.by_sample
        paint_network_overtime_up_down(
            graph,
            list_up=up_measured_list,
            list_down=down_measured_list,
            save_name=prefix,
            color_down=color_down,
            color_up=color_up, labels=labels
        )

    def _check_node(self, node_list):
        """
        Checks to see if list of nodes is in the graph, removes them if not.
        Parameters
        ----------
        node_list : list_like

        Returns
        -------
        set
        """
        node_list = set(node_list)
        missing_nodes = set()
        for i in node_list:
            if i not in self.nodes:
                missing_nodes.add(i)
        if len(missing_nodes):
            print("Warning : {} nodes do not exist in graph\n"
                  "Removing from list".format(len(missing_nodes)))

        return sorted(node_list.difference(missing_nodes))

    def _list_paths_to_graph(self, paths):
        graph = nx.DiGraph()
        current_nodes = set(graph.nodes())
        _edges = set()

        def _add_node(node):
            if node not in current_nodes:
                graph.add_node(node, **self.network.node[node])
                current_nodes.add(node)

        def _add_edge(i, j):
            if (i, j) not in _edges:
                graph.add_edge(i, j, **self.network.edges[i, j])
                _edges.add((i, j))

        # flatten paths into list of lists
        paths = [n for x in paths for n in x if n]

        for path in paths:
            previous = path[0]
            _add_node(previous)
            for protein in path[1:]:
                _add_node(protein)
                _add_edge(previous, protein)
                previous = protein
        if not len(graph.nodes):
            raise Exception("Resulting graph has no Nodes")
        return graph

    @staticmethod
    def _include_only(network, include_list):
        if not isinstance(include_list, (list, set)):
            raise AssertionError("Include list must be a list of set")
        sg = network.copy()
        all_nodes = set(sg.nodes())
        not_found = all_nodes.difference(set(include_list))
        sg.remove_nodes_from(not_found)
        if len(sg.nodes()) == 0:
            print("Warning: no nodes were found in include_only list! "
                  "Network doesn't contain any nodes!")
        return sg

    @staticmethod
    def _save_or_draw(graph, save_name, draw, img_format='png'):
        nx.write_gml(graph, "{}.gml".format(save_name))
        if draw:
            draw_graphviz(graph, save_name=save_name, image_format=img_format)

    @staticmethod
    def _max_distance(path_list, max_dist):

        new_paths = []
        for p in path_list:
            local_p = []
            for i in p:
                if len(i) <= max_dist:
                    local_p.append(i)
            new_paths.append(local_p)
        return new_paths


def _find_nx_path(node, network, single_path):
    node1, node2 = node

    l1 = _nx_find_path(network, node1, node2, single_path)
    l2 = _nx_find_path(network, node2, node1, single_path)
    return l1 + l2


def _nx_find_path(network, node1, node2, single_path=False):
    try:
        if single_path:
            return [nx.shortest_path(network, node1, node2)]
        else:
            return [p for p in nx.all_shortest_paths(network, node1, node2)]

    except nx.NetworkXNoPath:
        return []


if __name__ == '__main__':
    net = nx.DiGraph()
    net.add_node('X', label='XX', db='test1')
    net.add_node('Y', label='YY', db='test2')
    net.add_edge('X', 'B', intType='here')
    net.add_edge('X', 'C', intType='here')
    net.add_edge('B', 'Y', intType='here')
    net.add_edge('C', 'Y', intType='here')
    net.add_edge('D', 'Y', intType='here')
    net.add_edge('R', 'D', intType='here')
    net.add_edge('X', 'R', intType='here')

    net = nx.read_gpickle('background_network.p.gz')

    net = net.subgraph(
        net.neighbors('BAX') + ['BAX'] + net.neighbors('TP53') + ['TP53'])
    # print(len(net.edges()))
    # quit()
    # print(x.neighbors('X').nodes())
    # print(x.downstream_of_node('D').nodes())

    # print(x.path_between_2('X', ['Y', 'B', 'C']))
    # quit()

    x = Subgraph(net)
    g = x.paths_between_list(['X', 'Y', 'B', 'D'], single_path=True,
                             draw=False, save_name='test')

    print(g.nodes())
    # g = x.paths_between_list(['X', 'Y'], single_path=False,  draw=True, save_name='test')
    # print(g.nodes(data=True))




# from magine.networks.databases import load_kegg, load_biogrid_network, \
#     load_signor, load_reactome_fi, load_hmdb_network
# from magine.networks.databases.biogrid_interactions import BioGridDownload
#
# fresh_download = True
# verbose = True
#
#
# def test_download():
#     bgn = BioGridDownload()
#     bgn.parse_network()
#
#
# def test_kegg():
#     kegg_network = load_kegg(fresh_download=fresh_download,
#                                     verbose=verbose)
#
#
# def test_hmbd():
#     hmdb_network = load_hmdb_network(fresh_download=fresh_download,
#                                      verbose=verbose)
#
#
# def test_biogrid():
#     biogrid_network = load_biogrid_network(fresh_download=fresh_download,
#                                            verbose=verbose)
#
#
# def test_signor():
#     signor_network = load_signor(fresh_download=fresh_download,
#                                  verbose=verbose)
#
#
# def test_reactome():
#     reactome_network = load_reactome_fi(verbose=verbose)



import warnings
from itertools import chain

import matplotlib.pyplot as plt
import scipy.cluster.hierarchy as sch
import seaborn as sns


def heatmap_from_array(data, convert_to_log=False, y_tick_labels='auto',
                       cluster_row=False, cluster_col=False,
                       columns='sample_id', index='term_name',
                       values='combined_score', div_colors=False, num_colors=7,
                       figsize=(6, 4), sort_row=None, annotate_sig=False,
                       rank_index=None,
                       linewidths=0.0, cluster_by_set=False, min_sig=0):
    """

    Parameters
    ----------
    data : magine.data.base.BaseData
    convert_to_log : bool
        Convert fold_change column to log2 scale
    y_tick_labels : list_like
    columns : str
        Name of columns of df for pivot
    index : str
        Name of index of df for pivot
    values : str
        Name of values of df for pivot
    cluster_col : bool
        Cluster the data using searborn.clustermap
    cluster_row : bool
        Cluster the data using searborn.clustermap
    div_colors : bool
        Use divergent colors for plotting
    figsize : tuple
        Size of figure, passed to matplotlib/seaborn
    sort_row : str
        Sort rows by ('index', 'mean', max')
    num_colors : int
        Number of colors for color bar
    annotate_sig : bool
        Add '*' annotation to plot for significant changed terms
    linewidths : float or None
        Add white line between plots
    cluster_by_set: bool
        Cluster by gene set column. Only works for enrichment_array
    min_sig : int
        Minimum number of significant 'index' across samples. Can be used to
        remove rows that are not significant across any sample.
    rank_index: bool
        Rank rows by index. Deprecated , plus use sort_row arg instead.
    Returns
    -------
    plt.Figure

    """
    if min_sig:
        d_copy = data.require_n_sig(columns=columns, index=index,
                                    n_sig=min_sig)
    else:
        d_copy = data.copy()
    array = d_copy.pivoter(convert_to_log, columns=columns, index=index,
                           fill_value=0.0, values=values, min_sig=min_sig)
    if not len(array):
        warnings.warn("Empty array after filtering.")
        return
    # default values to be overwritten below
    col_colors = None
    annotations = None
    col_color_map = None
    col_labels = None
    fmt = None
    linkage = None
    add_col_group = False

    if rank_index is not None:
        warnings.warn("rank_index is deprecated; use sort_row='index'",
                      DeprecationWarning)
        if rank_index:
            # assuming provided true, will sort by rank.
            array.sort_index(ascending=True, inplace=True)

    if sort_row is not None:
        if sort_row not in ('index', 'max', 'mean', 'min'):
            raise ValueError("Can sort rows by 'index' name or 'max', 'min',"
                             "'mean' of values")

    # rank by index or cluster by term column
    if sort_row == 'index':
        array.sort_index(ascending=True, inplace=True)
    elif sort_row == 'mean':
        new_index = array.mean(axis=1).sort_values(ascending=False).index
        array = array.reindex(new_index)
    elif sort_row == 'max':
        new_index = array.max(axis=1).sort_values(ascending=False).index
        array = array.reindex(new_index)
    elif sort_row == 'min':
        new_index = array.max(axis=1).sort_values(ascending=False).index
        array = array.reindex(new_index)
    if cluster_by_set and "genes" in d_copy.columns:
        # clustering will be based on jaccard index of terms
        dist_mat, names = d_copy.calc_dist(level='sample')
        linkage = sch.linkage(dist_mat, method='average')
        # Add row cluster flag in case user didn't set
        cluster_row = True
        array = array.reindex(names)

    # set coloring scheme for heatmap
    if div_colors:
        pal = sns.color_palette("coolwarm", num_colors)
        center = 0
    else:
        pal = sns.light_palette("purple", as_cmap=True)
        center = None

    # Group together by columns if provided
    if isinstance(columns, (list, tuple)) and len(columns) == 2:
        add_col_group = True
        col_labels, col_colors, col_color_map = _set_col_colors(array)

    # check annotations exist
    if annotate_sig:
        annotate_sig, annotations, fmt = _get_sig_annotations(array, d_copy,
                                                              columns,
                                                              index, min_sig)
    cluster_args = dict(method='complete', metric='correlation')
    if cluster_row or cluster_col or add_col_group:
        fig = sns.clustermap(array, cmap=pal, center=center,
                             yticklabels=y_tick_labels,
                             col_colors=col_colors,
                             col_cluster=cluster_col,
                             row_cluster=cluster_row,
                             row_linkage=linkage,
                             figsize=figsize, linewidths=linewidths,
                             annot=annotations, fmt=fmt,
                             **cluster_args
                             )

        # We need to reorder the annotations if we cluster
        if annotate_sig:
            if cluster_row:
                annotations = annotations[fig.dendrogram_row.reordered_ind]
            if cluster_col:
                annotations = annotations[:, fig.dendrogram_col.reordered_ind]
            # Only need figure for dendrogram ordering, not actual plot.
            # Can probably do this without the plotting interface, but this
            # seems to do the job for now.

            plt.close()

            # make final figure
            fig = sns.clustermap(array, cmap=pal, center=center,
                                 yticklabels=y_tick_labels,
                                 col_colors=col_colors,
                                 col_cluster=cluster_col,
                                 row_cluster=cluster_row,
                                 row_linkage=linkage,
                                 figsize=figsize, linewidths=linewidths,
                                 annot=annotations, fmt=fmt,
                                 **cluster_args
                                 )

        # add labels to column colors
        if add_col_group:
            fig = _add_column_color_groups(d_copy, fig, col_color_map,
                                           col_labels, columns)

        # add clustered columns and rows, basically allows us to extract out
        # the clusters from the figure, if we wanted to do something with them
        # ie run enrichment analysis.
        if cluster_col:
            col_cltrs = sch.fcluster(fig.dendrogram_col.linkage, t=2,
                                     criterion='maxclust')
            col_cltrs = col_cltrs[fig.dendrogram_col.reordered_ind]
            col_clusters = dict()
            for i in sorted(set(col_cltrs)):
                cols = fig.data2d.columns.values[col_cltrs == i]
                col_clusters[i] = fig.data2d[cols]
            fig.col_clusters = col_clusters
        if cluster_row:
            row_cltrs = sch.fcluster(fig.dendrogram_row.linkage, t=2,
                                     criterion='maxclust')
            row_cltrs = row_cltrs[fig.dendrogram_row.reordered_ind]
            row_clusters = dict()
            for i in sorted(set(row_cltrs)):
                row_clusters[i] = fig.data2d.loc[row_cltrs == i].index.values
            fig.row_clusters = row_clusters
        fig.ax_heatmap.set_ylabel('')
        fig.ax_heatmap.set_xlabel('')
        # plt.subplots_adjust(right=0.7, top=1.5)
    else:
        fig = plt.figure(figsize=figsize)
        ax = fig.add_subplot(111)
        sns.heatmap(array, ax=ax, yticklabels=y_tick_labels, cmap=pal,
                    center=center, annot=annotations, fmt=fmt,
                    linewidths=linewidths)
        ax.set_ylabel('')
        ax.set_xlabel('')
    return fig


def heatmap_by_terms(data, term_labels, term_sets, colors=None, min_sig=None,
                     convert_to_log=False, y_tick_labels='auto',
                     columns='sample_id', index='identifier',
                     values='fold_change', linewidths=0,
                     cluster_row=False, cluster_col=False, div_colors=False,
                     num_colors=21, figsize=None, annotate_sig=False,
                     **kwargs):
    """

    Parameters
    ----------
    data : pd.DataFrame
    term_labels : list_like
        List of labels for grouping
    term_sets : list_like
        List of list like that create the terms
    colors : list_like
        Colors for plotting, if not provided it will be created
    min_sig : int
        Number of sign
    convert_to_log : bool
    y_tick_labels : list_like
    columns : str
        Name of columns of df for pivotn
    index : str
        Name of index of df for pivot
    values : str
        Name of values of df for pivot
    cluster_col : bool
        Cluster the data using searborn.clustermap
    cluster_row : bool
        Cluster rows
    div_colors : bool
        Use divergent colors for plotting
    figsize : tuple
        Size of figure, passed to matplotlib/seaborn
    num_colors : int
        Number of colors for color bar
    annotate_sig : bool
        Add '*' annotation to plot for significant changed terms
    linewidths : float or None
        Add white line between plots
    min_sig : int
        Minimum number of significant 'index' across samples. Can be used to
        remove rows that are not significant across any sample.

    Returns
    -------
    plt.Figure

    """

    if len(term_labels) != len(term_sets):
        raise AssertionError("Number of term_labels must "
                             "equal number of term_sets")
    # default values to be overwritten below
    annotations = None
    fmt = None
    add_col_group = False
    tmp_d = data.copy()
    if index == 'label':
        id_to_label = dict()
        for i, j in tmp_d[['identifier', 'label']].values:
            if i not in id_to_label:
                id_to_label[i] = set()
            id_to_label[i].add(j)
        new_term_sets = [
            set(chain.from_iterable([id_to_label[j] for j in i
                                     if j in id_to_label]))
            for i in term_sets
        ]
    else:
        new_term_sets = term_sets
    all_items = set(chain.from_iterable(new_term_sets))
    tmp_d = tmp_d.loc[tmp_d[index].isin(all_items)]

    # pivot datatable
    array = tmp_d.pivoter(convert_to_log, columns=columns, index=index,
                          fill_value=0.0, values=values, min_sig=min_sig)
    if not len(array):
        warnings.warn("Empty array after filtering.")
        return
    if colors is None:
        colors = sns.color_palette("Paired", n_colors=len(term_labels))
    else:
        if len(colors) != len(new_term_sets):
            raise AssertionError("Number of colors must "
                                 "equal number of term_labels")
    vals = set(array.index.values)
    final_sorted = sorted(new_term_sets[0].intersection(vals))
    added = set(final_sorted)
    # create colors for each

    row_colors = [colors[0] for _ in added]
    to_remove = set()
    for term, color, cname in zip(new_term_sets[1:], colors[1:],
                                  term_labels[1:]):
        added_any = False
        for i in sorted(term.intersection(vals)):
            if i not in added:
                added_any = True
                row_colors.append(color)
                final_sorted.append(i)
                added.add(i)
        if not added_any:
            to_remove.add(cname)
    # only keep indexes that are in the provided sets
    array = array[array.index.isin(final_sorted)]

    # resort according to color
    array = array.reindex(final_sorted)
    if isinstance(columns, list) and len(columns) == 2:
        add_col_group = True
        col_labels, col_colors, col_color_map = _set_col_colors(array)
    else:
        col_labels, col_colors, col_color_map = None, None, None
    # set colors map for heatmap
    if div_colors:
        pal = sns.color_palette("coolwarm", num_colors)
        center = 0
    else:
        pal = sns.light_palette("red", n_colors=len(new_term_sets),
                                as_cmap=True)
        center = None

    if annotate_sig:
        annotate_sig, annotations, fmt = _get_sig_annotations(array, data,
                                                              columns,
                                                              index, min_sig)
    cluster_args = dict(method='single', metric='correlation')
    fig = sns.clustermap(array,
                         yticklabels=y_tick_labels,
                         figsize=figsize, linewidths=linewidths,
                         row_colors=row_colors, col_colors=col_colors,
                         col_cluster=cluster_col, row_cluster=cluster_row,
                         cmap=pal, center=center,
                         annot=annotations, fmt=fmt,
                         **cluster_args
                         )
    if annotate_sig:
        if cluster_col:
            annotations = annotations[:, fig.dendrogram_col.reordered_ind]
        if cluster_row:
            annotations = annotations[fig.dendrogram_row.reordered_ind]
        plt.close()
        fig = sns.clustermap(array,
                             yticklabels=y_tick_labels,
                             figsize=figsize, linewidths=linewidths,
                             row_colors=row_colors, col_colors=col_colors,
                             col_cluster=cluster_col, row_cluster=cluster_row,
                             cmap=pal, center=center,
                             annot=annotations, fmt=fmt,
                             **cluster_args
                             )
    for color, label in zip(colors, term_labels):
        if label in to_remove:
            continue
        fig.ax_row_dendrogram.bar(0, 0, color=color, label=label, linewidth=0)
    fig.ax_row_dendrogram.legend(loc=0, ncol=1)
    if add_col_group:
        fig = _add_column_color_groups(tmp_d, fig, col_color_map, col_labels,
                                       columns)
    fig.ax_heatmap.set_ylabel('')
    fig.ax_heatmap.set_xlabel('')
    plt.subplots_adjust(right=0.7)
    return fig


def cluster_distance_mat(dist_mat, names, figsize=(8, 8)):
    """ Creates heatmap from distance matrix.

    Parameters
    ----------
    dist_mat : np.array
        Distance matrix array.
    names : list_like
        Names of ticks for distance matrix
    figsize : tuple
        Size of figure, passed to matplotlib

    Returns
    -------

    """
    # Compute and plot first dendrogram.
    fig = plt.figure(figsize=figsize)

    # Compute and plot second dendrogram.
    ax2 = fig.add_axes([0.3, 0.71, 0.6, 0.2])
    Y = sch.linkage(dist_mat, method='average')
    Z2 = sch.dendrogram(Y)
    ax2.set_xticks([])
    ax2.set_yticks([])

    # Plot distance matrix.
    axmatrix = fig.add_axes([0.3, 0.1, 0.6, 0.6])

    # reorder matrix
    idx1 = Z2['leaves']
    dist_mat = dist_mat[idx1, :]
    dist_mat = dist_mat[:, idx1]
    names = names[idx1]

    # create figure
    im = axmatrix.matshow(dist_mat, aspect='auto', origin='lower',
                          cmap=plt.cm.Reds, vmin=0, vmax=1)

    # add xtick labels
    axmatrix.set_xticks(range(len(names)))
    axmatrix.set_xticklabels(names, minor=False)
    axmatrix.xaxis.set_label_position('bottom')
    axmatrix.xaxis.tick_bottom()
    plt.xticks(rotation=90, fontsize=8)

    # add ytick labels
    axmatrix.set_yticks(range(len(names)))
    axmatrix.set_yticklabels(names, minor=False)
    axmatrix.yaxis.set_label_position('left')
    axmatrix.yaxis.tick_left()
    plt.yticks(rotation=0, fontsize=8)

    # add colorbar
    axcolor = fig.add_axes([0.94, 0.1, 0.02, 0.6])
    plt.colorbar(im, cax=axcolor)

    return fig


def _set_col_colors(array):
    col_labels = array.columns.levels[0]
    labels = list(array.columns.levels[1])
    col_color_map = sns.color_palette("Dark2", len(col_labels))
    col_colors = [col_color_map[i] for i in array.columns.codes[0]]
    array.columns = [labels[i] for i in array.columns.codes[1]]
    return col_labels, col_colors, col_color_map


def _add_column_color_groups(data, fig, colors, color_labels, columns):
    for color, label in zip(colors, color_labels):
        fig.ax_col_dendrogram.bar(0, 0, color=color, label=label, linewidth=0)
    plt.setp(fig.ax_col_dendrogram.yaxis.get_majorticklabels(), rotation=0,
             fontsize=16)
    fig.ax_col_dendrogram.legend(loc="center", fontsize=12, ncol=2,
                                 bbox_to_anchor=(0.5, 1., 0.5, 0.5))
    v_line_list = []
    prev = 0
    for i in color_labels:
        n_samples = len(data[data[columns[0]] == i][columns[1]].unique())
        prev += n_samples
        v_line_list.append(prev)

    fig.fig.axes[2].vlines(v_line_list, *fig.fig.axes[2].get_ylim())
    fig.fig.axes[3].vlines(v_line_list, *fig.fig.axes[3].get_ylim())
    return fig


def _get_sig_annotations(arr, dat, columns, index, min_sig):
    # Have to rank by column for this to work
    if 'significant' in dat.columns:
        tmp2 = dat.pivoter(False, columns=columns, index=index,
                           values='significant', fill_value=0, min_sig=min_sig)
        tmp2 = tmp2.reindex(arr.index)
        tmp2[tmp2 > 0] = True
        tmp2 = tmp2.replace(0, '')
        tmp2 = tmp2.replace(False, '')
        tmp2 = tmp2.replace(True, '+')
        return True, tmp2.values, ''
    else:
        print("To annotate please add a significant column to data")
        return False, None, None



import types

import matplotlib.pyplot as plt
import pandas as pd
from wordcloud import STOPWORDS, WordCloud

# Will be OK in Python 2
try:
    basestring
# Allows isinstance(foo, basestring) to work in Python 3
except:
    basestring = str

process_dbs = [
    'GO_Biological_Process_2017',
    'Humancyc_2016',
    'Reactome_2016',
    'KEGG_2016',
    'NCI-Nature_2016',
    'Panther_2016',
    'WikiPathways_2016',
]

basic_words = {'signaling', 'signalling', 'receptor', 'events', 'protein',
               'proteins', 'regulation', 'interactions', 'via', 'signal',
               'mediated', 'pathway', 'activity', 'complex', 'positive',
               'mrna', 'cellular', 'viral', 'host', 'processing', 'activation',
               'rrna', 'network', 'rna', 'cancer', 'disease', 'cascade',
               'transcript', 'influenza', 'beta', 'pathways', 'gene', 'hiv',
               'downstream', 'activated', 'target', 'dn', 'up'
               }

basic_words.update(STOPWORDS)


def word_cloud_from_array(enrichment_array, sample_ids, category=None,
                          database_list=None, save_name=None, p_value=0.05):
    """
    Creates a word cloud for each sample_id


    Parameters
    ----------
    enrichment_array : magine.enrichment.tools.EnrichmentResults
    sample_ids : list
    category : str, list, optional
    database_list : str, list, optional
    save_name : str

    Returns
    -------

    """
    samples = []
    all_samples = []
    for i in sample_ids:
        sample = enrichment_array.filter_multi(
            p_value=p_value, db=database_list, sample_id=i, category=category
        )

        all_samples.append(sample)
        if len(sample) != 0:
            if save_name is not None:
                output = "{}_{}_wordcloud".format(save_name, i)
            else:
                output = None
            hits_1 = create_wordcloud(sample, save_name=output)
            df1 = pd.DataFrame(
                list(hits_1.word_dict.items()), columns=['words', 'counts']
            )
            df1.sort_values('counts', ascending=False, inplace=True)
            df1['sample'] = i
            samples.append(df1.copy())

    df = pd.concat(samples)
    samples = df['sample'].unique()
    df = pd.pivot_table(df, index='words', columns=['sample']).fillna(0)
    df.columns = df.columns.droplevel()
    df['sum'] = df[samples].sum(axis=1)
    df.sort_values('sum', ascending=False, inplace=True)
    # print("\nSorted by sum of all")
    # print(df.sort_values('sum', ascending=False).head(25))
    return all_samples, df


def create_wordcloud(df, save_name=None):
    """
    Creates a word cloud based on enrichment array

    Must have column 'term_name'.
    It takes this column, flattens it into a large string.
    Then we pass it to the python package wordcloud, which generates the
    wordcloud.

    It returns the figure with a save method and a dictionary of counts.



    Parameters
    ----------
    df : pd.DataFrame
    save_name : str

    Returns
    -------

    """
    data = df.apply(_cleanup_term_name, axis=1)
    text = ' '.join(data)
    # Generate a word cloud image
    wc = WordCloud(margin=0, background_color=None, mode='RGBA',
                   # min_count=1,
                   width=800, height=600, collocations=True,
                   stopwords=basic_words)
    wordcloud = wc.generate(text)
    word_dict = wc.process_text(text)

    def plot(self, save_name=None, figsize=(8, 5)):
        fig = plt.figure(figsize=figsize)
        ax = fig.add_subplot(111)
        ax.imshow(self, interpolation='bilinear')
        plt.xticks([])
        plt.yticks([])
        plt.axis("off")
        if save_name is not None:
            plt.savefig('{}.png'.format(save_name), bbox_inches='tight',
                        dpi=150)
            plt.title(save_name)
        return fig

    wordcloud.plot = types.MethodType(plot, wordcloud)

    if save_name is not None:
        wordcloud.plot(save_name)

    wordcloud.word_dict = word_dict

    df1 = pd.DataFrame(list(word_dict.items()),
                       columns=['words', 'counts'])

    df1.sort_values('counts', ascending=False, inplace=True)
    wordcloud.data = df1

    return wordcloud


def _cleanup_term_name(row):
    if not isinstance(row['term_name'], basestring):
        print(row)
    x = row['term_name'].split('_')[0].lower()
    x = ' ' + x
    x = x.replace(' p53', ' tp53')
    x = x.replace('  ', ' ')
    if x[0] == ' ':
        x = x[1:]
    if x[-1] == ' ':
        x = x[:-1]
    return x



import os
import subprocess
import warnings

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from pandas.plotting import table

from magine.data.base import BaseData
from magine.data.tools import log2_normalize_df
from magine.plotting import volcano_plots as v_plot
from magine.plotting.species_plotting import plot_dataframe, plot_species

# pandas.set_option('display.max_colwidth', -1)
# column definitions

fold_change = 'fold_change'
flag = 'significant'
exp_method = 'source'
p_val = 'p_value'
rna = 'rna_seq'
gene = 'gene'
protein = 'protein'
metabolites = 'metabolites'
species_type = 'species_type'
sample_id = 'sample_id'
identifier = 'identifier'
label = 'label'
valid_cols = [fold_change, flag, p_val, species_type, sample_id]


def load_data_csv(file_name, **kwargs):
    """ Deprecated; use :class:`load_data` instead. """

    warnings.warn("load_data_csv will be removed in a future "
                  "version of MAGINE. Use load_data instead.",
                  DeprecationWarning, stacklevel=2)
    return load_data(file_name, **kwargs)


def load_data(file_name, **kwargs):
    """ Load data into EnrichmentResult data class

    Parameters
    ----------
    file_name : str
    kwargs :
        Flags to pass to pandas.

    Returns
    -------
    df : EnrichmentResult

    """
    df = pd.read_csv(file_name, **kwargs)
    df = df[df[fold_change].notnull()]
    return ExperimentalData(df)


class Sample(BaseData):
    """ Provides tools for subsets of data types

    """

    def __init__(self, *args, **kwargs):
        super(Sample, self).__init__(*args, **kwargs)
        # self.drop_duplicates(inplace=True)
        self._index = identifier
        self._identifier = identifier
        self._value_name = fold_change
        self._sample_id_name = sample_id
        self._label = label
        self._up = None
        self._down = None
        self._sig = None

    @property
    def _constructor(self):
        return Sample

    @property
    def exp_methods(self):
        """ List of sample_ids in data"""
        return sorted(set(self[exp_method].values))

    @property
    def sample_ids(self):
        """ List of sample_ids in data"""
        return sorted(set(self[sample_id].values))

    @property
    def up(self):
        """return up regulated species"""
        return self.loc[self[flag] & (self[fold_change] > 0)]

    @property
    def down(self):
        """return down regulated species"""
        return self.loc[self[flag] & (self[fold_change] < 0)]

    @property
    def id_list(self):
        """ Set of species identifiers """
        return set(self[self._identifier].values)

    @property
    def label_list(self):
        """ Set of species labels """
        return set(self[self._label].values)

    @property
    def up_by_sample(self):
        """List of up regulated species by sample"""
        return [self.loc[self[sample_id] == i].up.id_list
                for i in self.sample_ids]

    @property
    def down_by_sample(self):
        """List of down regulated species by sample"""
        return [self.loc[self[sample_id] == i].down.id_list
                for i in self.sample_ids]

    @property
    def by_sample(self):
        """List of significantly flagged species by sample"""
        return [self.loc[self[sample_id] == i].id_list
                for i in self.sample_ids]

    def subset(self, species=None, index='identifier', sample_ids=None,
               exp_methods=None):
        """

        Parameters
        ----------
        species : list, str
            List of species to create subset dataframe from
        index : str
            Index to filter based on provided 'species' list
        sample_ids : str, list
            List or string to filter sample
        exp_methods : str, list
            List or string to filter sample

        Returns
        -------
        magine.data.experimental_data.Species
        """
        df = self.copy()
        if isinstance(species, str):
            df = df.loc[df[index].str.contains(species)]
        elif isinstance(species, (list, tuple, set)):
            df = df.loc[df[index].isin(species)]
        if sample_ids is not None:
            if isinstance(species, str):
                df = df.loc[df[sample_id].str.contains(sample_ids)]
            else:
                df = df.loc[df[sample_id].isin(sample_ids)]
        if exp_methods is not None:
            if isinstance(species, str):
                df = df.loc[df[exp_method].str.contains(exp_methods)]
            else:
                df = df.loc[df[exp_method].isin(exp_methods)]
        return df

    def plot_pie_sig_ratio(self, save_name=None, ax=None, fig=None,
                           figsize=None):
        """

        Parameters
        ----------
        save_name : str
        ax : matplotlib.axes, optional
        fig : matplotlib.figure
        figsize : tuple
            Size of figure

        Returns
        -------

        """
        x = len(self.id_list)
        y = len(self.sig.id_list)
        total = x + y
        if fig is None and ax is None:
            if figsize is None:
                figsize = (3, 3)
            fig = plt.figure(figsize=figsize)
            ax = fig.add_subplot(111)
        wedges, texts, autotexts = ax.pie(
            [x, y],
            explode=(0.05, 0.05),
            textprops={'fontsize': 16},
            autopct=lambda p: '{:.0f}'.format(p * total / 100),
            shadow=True,
            startangle=140
        )

        plt.setp(autotexts, size=20)
        plt.axis('equal')
        if save_name is not None:
            plt.savefig('{}.png'.format(save_name), dpi=300,
                        bbox_inches='tight')
        return fig

    def volcano_plot(self, save_name=None, out_dir=None, sig_column=False,
                     p_value=0.1, fold_change_cutoff=1.5, x_range=None,
                     y_range=None):
        """ Create a volcano plot of data


        Parameters
        ----------
        save_name: str
            name to save figure
        out_dir: str, directory
            Location to save figure
        sig_column: bool, optional
            If to use significant flags of data
        p_value: float, optional
            Criteria for significant
        fold_change_cutoff: float, optional
            Criteria for significant
        y_range: array_like
            upper and lower bounds of plot in y direction
        x_range: array_like
            upper and lower bounds of plot in x direction

        Returns
        -------
        matplotlib.Figure

        """
        fig = v_plot.volcano_plot(self, save_name=save_name, out_dir=out_dir,
                                  sig_column=sig_column, p_value=p_value,
                                  fold_change_cutoff=fold_change_cutoff,
                                  x_range=x_range, y_range=y_range)
        return fig

    def volcano_by_sample(self, save_name=None, p_value=0.1,
                          out_dir=None, fold_change_cutoff=1.5, y_range=None,
                          x_range=None, sig_column=False):
        """
        Creates a figure of subplots of provided experimental method

        Parameters
        ----------
        save_name: str
            name to save figure
        out_dir: str, directory
            Location to save figure
        sig_column: bool, optional
            If to use significant flags of data
        p_value: float, optional
            Criteria for significant
        fold_change_cutoff: float, optional
            Criteria for significant
        y_range: array_like
            upper and lower bounds of plot in y direction
        x_range: array_like
            upper and lower bounds of plot in x direction

        Returns
        -------

        """

        data = self.copy()
        n_sample = np.sort(data[sample_id].unique())

        if len(n_sample) > 8:
            n_cols = 3
        else:
            n_cols = 2
        n_rows = np.rint(np.rint(len(n_sample) / float(n_cols)))
        if n_cols * n_rows < len(n_sample):
            if n_cols >= n_rows:
                n_rows += 1
            else:
                n_cols += 1

        fig = plt.figure(figsize=(4 * n_rows, 3 * n_cols))
        for n, i in enumerate(n_sample):
            sample = data[data[sample_id] == i].copy()

            sample = sample.dropna(subset=[p_val])
            sample = sample[np.isfinite(sample[fold_change])]
            sample = sample.dropna(subset=[fold_change])
            sec_0, sec_1, sec_2 = v_plot.create_mask(sample, sig_column,
                                                     p_value,
                                                     fold_change_cutoff)

            ax = fig.add_subplot(n_rows, n_cols, n + 1)
            ax.set_title(i)
            v_plot.add_volcano_plot(ax, sec_0, sec_1, sec_2)
            if not sig_column:
                fc = np.log2(fold_change_cutoff)
                log_p_val = -1 * np.log10(p_value)
                ax.axvline(x=fc, linestyle='--')
                ax.axvline(x=-1 * fc, linestyle='--')
                ax.axhline(y=log_p_val, linestyle='--')
            if y_range is not None:
                ax.set_ylim(y_range[0], y_range[1])
            if x_range is not None:
                ax.set_xlim(x_range[0], x_range[1])
        fig.tight_layout()
        if save_name is not None:
            v_plot.save_plot(fig, save_name=save_name, out_dir=out_dir)
        return fig

    def plot_species(self, species_list=None, subset_index=None,
                     save_name=None, out_dir=None, title=None,
                     plot_type='plotly', image_format='png'):
        """
        Create scatter plot of species list

        Parameters
        ----------
        species_list : list
            list of compounds
        subset_index : list
            Column to filter based on species_list
        save_name : str
            Name of html output file
        out_dir : str
            Location to place plots
        title : str
            Title for HTML page
        plot_type : str
            Type of plot outputs, can be "plotly" or "matplotlib"
        image_format : str
            pdf or png, only used if plot_type="matplotlib"

        Returns
        -------
        matplotlib.Figure or plotly.Figure
        """
        df = self.copy()
        if species_list is not None:
            if subset_index is None:
                subset_index = self._index
            df = df.subset(species_list, index=subset_index)
        return plot_species(
            df, save_name=save_name, out_dir=out_dir, title=title,
            plot_type=plot_type, image_format=image_format
        )

    def plot_all(self, html_file_name, out_dir='out',
                 plot_type='plotly', run_parallel=False):
        """
        Creates a plot of all metabolites

        Parameters
        ----------
        html_file_name : str
            filename to save html of all plots
        out_dir: str, path
            Directory that will contain all proteins
        plot_type : str
            plotly or matplotlib output
        run_parallel : bool
            Create the plots in parallel
        Returns
        -------

        """

        plot_dataframe(self, html_filename=html_file_name,
                       out_dir=out_dir, plot_type=plot_type,
                       run_parallel=run_parallel)

    def plot_histogram(self, save_name=None, y_range=None, out_dir=None):
        """
        Plots a histogram of data

        Parameters
        ----------
        save_name: str
            Name of figure
        out_dir: str, path
            Path to location to save figure
        y_range: array_like
            range of data


        Returns
        -------

        """
        data = self.copy()
        data = data.dropna(subset=[p_val])
        data = data[np.isfinite(data[fold_change])]
        data = data.dropna(subset=[fold_change])

        tmp = np.array(log2_normalize_df(data, fold_change)[fold_change])

        fig = plt.figure()
        ax = fig.add_subplot(111)
        ax.hist(tmp, 50, color='gray')
        if y_range is not None:
            plt.xlim(y_range[0], y_range[1])

        ax.set_yscale('log', basey=10)
        ax.set_xlabel('log$_2$ Fold Change', fontsize=16)
        ax.set_ylabel('Count', fontsize=16)
        fig.tight_layout()
        if save_name is not None:
            v_plot.save_plot(fig, save_name, out_dir)
        return fig


class ExperimentalData(object):
    """
    Manages all experimental data

    """

    def __init__(self, data_file):
        """

        Parameters
        ----------
        data_file : str, pandas.DataFrame
            Name of file, generally csv.
            If provided a str, the file will be read in as a pandas.DataFrame


        """
        if isinstance(data_file, pd.DataFrame):
            df = data_file.copy()
        else:
            df = pd.read_csv(data_file, parse_dates=False, low_memory=False)
        df.reset_index(drop=True, inplace=True)
        df.drop_duplicates(inplace=True)
        for i in valid_cols:
            if i not in df.dtypes:
                print("{} not in columns.".format(i))

        self.data = BaseData(df)
        self._index = 'identifier'
        self.__proteins = None
        self.__genes = None
        self.__species = None
        self.__rna = None
        self.__compounds = None
        for i in self.exp_methods:
            self.__setattr__(i, Sample(
                self.data.loc[self.data[exp_method] == i]))
        for i in self.sample_ids:
            self.__setattr__(i, Sample(
                self.data.loc[self.data[sample_id] == i]))

    def __setattr__(self, name, value):
        super(ExperimentalData, self).__setattr__(name, value)

    def __getitem__(self, name):
        return super(ExperimentalData, self).__getattribute__(name)

    @property
    def genes(self):
        """ All data tagged with gene

        Includes protein and RNA.

        Returns
        -------

        """
        if self.__genes is None:
            tmp = self.data.copy()
            tmp = tmp.loc[tmp[species_type] == protein]
            self.__genes = Sample(tmp)
        return self.__genes

    @property
    def proteins(self):
        """ Protein level data

        Tagged with "gene" identifier that is not RNA

        Returns
        -------

        """
        if self.__proteins is None:
            tmp = self.data.copy()
            tmp = tmp.loc[(self.data[species_type] == protein) &
                          ~(tmp[exp_method] == rna)]
            self.__proteins = Sample(tmp)
        return self.__proteins

    @property
    def rna(self):
        """ RNA level data

        Tagged with "RNA"

        Returns
        -------

        """
        if self.__rna is None:
            tmp = self.data.copy()
            tmp = tmp.loc[tmp[exp_method] == rna]
            self.__rna = Sample(tmp)
        return self.__rna

    @property
    def compounds(self):
        """ Only compounds in data

        Returns
        -------
        Sample

        """
        if self.__compounds is None:
            tmp = self.data.copy()
            tmp = tmp.loc[tmp[species_type] == metabolites]
            self.__compounds = Sample(tmp)
        return self.__compounds

    @property
    def species(self):
        """ Returns data in Sample format

        Returns
        -------
        Sample

        """
        if self.__species is None:
            self.__species = Sample(self.data.copy())
        return self.__species

    @property
    def exp_methods(self):
        """ List of source columns """
        return list(self.data[exp_method].unique())

    @property
    def sample_ids(self):
        """ List of sample_ids """
        return sorted(list(self.data[sample_id].unique()))

    def subset(self, species, index='identifier'):
        """

        Parameters
        ----------
        species : list, str
            List of species to create subset dataframe from
        index : str
            Index to filter based on provided 'species' list

        Returns
        -------
        magine.data.experimental_data.Species
        """
        df = self.species.copy()
        if isinstance(species, str):
            df = df.loc[df[index].str.contains(species)]
        else:
            df = df.loc[df[index].isin(species)]
        return df

    def get_measured_by_datatype(self):
        """
        Returns dict of species per data type

        Returns
        -------
        dict

        """
        return get_measured_by_datatype(self)

    def create_summary_table(self, sig=False, index=identifier, save_name=None,
                             plot=False, write_latex=False):
        """
        Creates a summary table of data.


        Parameters
        ----------
        sig: bool
            Flag to summarize significant species only
        save_name: str
            Name to save csv and .tex file
        index: str
           Index for counts
        plot: bool
            If you want to create a plot of the table
        write_latex: bool
            Create latex file of table


        Returns
        -------
        pandas.DataFrame

        """
        return create_table_of_data(self, sig=sig, index=index,
                                    save_name=save_name, plot=plot,
                                    write_latex=write_latex)

    def volcano_analysis(self, out_dir, use_sig_flag=True,
                         p_value=0.1, fold_change_cutoff=1.5):
        """
        Creates a volcano plot for each experimental method

        Parameters
        ----------
        out_dir: str, path
            Path to where the output figures will be saved
        use_sig_flag: bool
            Use significant flag of data
        p_value: float, optional
            p value criteria for significant
            Will not be used if use_sig_flag
        fold_change_cutoff: float, optional
            fold change criteria for significant
            Will not be used if use_sig_flag

        Returns
        -------

        """
        if not os.path.exists(out_dir):
            os.mkdir(out_dir)
        for i in self.exp_methods:
            self[i].volcano_plot(
                i, out_dir=out_dir, sig_column=use_sig_flag,
                p_value=p_value, fold_change_cutoff=fold_change_cutoff
            )


template = r'''
\documentclass[12pt, letterpaper]{{article}}
\usepackage{{booktabs}}
\usepackage{{geometry}}
\usepackage{{pdflscape}}
%\usepackage{{nopageno}}
\geometry{{ papersize={{4.444in,12.681in}},total={{2.8in,6.8in}} }}
\begin{{document}}
\begin{{landscape}}
\begin{{table}}
{}
\end{{table}}
\end{{landscape}}
\end{{document}}
'''


def get_measured_by_datatype(data):
    """ Get unique list of species for each 'source' label in data.

    Parameters
    ----------
    data : ExperimentalData

    Returns
    -------
    measured, sig_measured : dict, dict
        Dictionaries where keys are 'source' and values are sets of ids.

    """

    measured = dict()
    sig_measured = dict()
    for i in data.exp_methods:
        sig_measured[i] = set(data[i].sig.id_list)
        measured[i] = set(data[i].id_list)
    return measured, sig_measured


def create_table_of_data(data, sig=False, index='identifier', save_name=None,
                         plot=False, write_latex=False):
    """
    Creates a summary table of data.


    Parameters
    ----------
    data : ExperimentalData
    sig: bool
        Flag to summarize significant species only
    save_name: None, str
        Name to save csv and .tex file
    index: str
        Index to create counts
    plot: bool
        If you want to create a plot of the table
    write_latex: bool
        Create latex file of table


    Returns
    -------
    pandas.DataFrame

    """

    if sig:
        data_copy = data.species.sig.copy()
    else:
        data_copy = data.species.copy()

    count_table = data_copy.pivot_table(values=index, index=exp_method,
                                        columns=sample_id, fill_value=np.nan,
                                        aggfunc=lambda x: x.dropna().nunique())

    # This just makes sure things are printed as ints, not floats
    for i in count_table.columns:
        count_table[i] = count_table[i].fillna(-1).astype(int).replace(-1, '-')
    unique_col = {}
    for i in data.exp_methods:
        if sig:
            unique_col[i] = len(set(data[i].sig[index].values))
        else:
            unique_col[i] = len(set(data[i][index].values))
    count_table['Total Unique Across'] = pd.Series(unique_col,
                                                   index=count_table.index)
    if plot:
        ax = plt.subplot(111, frame_on=False)

        table(ax, count_table, loc='center')
        ax.xaxis.set_visible(False)
        ax.yaxis.set_visible(False)
        plt.tight_layout()
        if save_name is not None:
            plt.savefig('{}.png'.format(save_name), dpi=300,
                        bbox_inches='tight')

    if save_name is not None:
        count_table.to_csv('{}.csv'.format(save_name))
    if write_latex and save_name is not None:
        _write_to_latex(pd_table=count_table, save_name=save_name)
    return count_table


def _write_to_latex(pd_table, save_name):
    filename = '{0}.tex'.format(save_name)

    with open(filename, 'wt') as f:
        st = pd_table.to_latex(
            column_format='*{{{}}}{{c}}'.format(str(pd_table.shape[1] + 2)))
        f.write(template.format(st))

    if _which('pdflatex'):
        print('Compiling table')
        with open(os.devnull, "w") as fnull:
            subprocess.call(['pdflatex', filename], stderr=subprocess.STDOUT,
                            stdout=fnull)

            if _which('pdfcrop'):
                pdffile = '{0}.pdf'.format(save_name)
                subprocess.call(['pdfcrop', pdffile, pdffile],
                                stderr=subprocess.STDOUT, stdout=fnull)
                if _which('convert'):
                    tmp_png_name = '{0}.png'.format(save_name)
                    subprocess.call(['convert', '-density', '300', pdffile,
                                     '-quality', '90', tmp_png_name],
                                    stderr=subprocess.STDOUT, stdout=fnull)
    else:
        print('Install pdflatex to compile to pdf or png\n'
              'You can use the csv file for use in outside tools')


def _which(program):
    def _is_exe(filepath):
        return os.path.isfile(filepath) and os.access(filepath, os.X_OK)

    fpath, fname = os.path.split(program)
    if fpath:
        if _is_exe(program):
            return program
        elif _is_exe(program + '.exe'):
            return program + '.exe'
    else:
        for path in os.environ["PATH"].split(os.pathsep):
            path = path.strip('"')
            exe_file = os.path.join(path, program)
            if _is_exe(exe_file):
                return exe_file

    return None



"""
GO analysis function using orange bioinformatics
"""

import os

import numpy as np
import pandas as pd
from statsmodels.stats.multitest import fdrcorrection
from statsmodels.stats.proportion import binom_test

from magine.data.storage import network_data_dir
from magine.enrichment.deprecated.databases.gene_ontology import \
    download_and_process_go
from magine.html_templates.html_tools import write_filter_table
from magine.plotting.species_plotting import plot_genes_by_ont

try:
    import cPickle as pickle
except:  # python3 doesnt have cPickle
    import pickle

pd.set_option('display.max_colwidth', -1)

evidence_codes = ['EXP', 'IDA', 'IPI', 'IMP', 'IGI', 'IEP', 'TAS', 'IC', 'IEA']


def pivot_table_for_export(data, save_name=None):
    """
    creates a pivot table of combined data that is in MAGINE format

    Parameters
    ----------
    data : pd.DataFrame
        output from magine analysis
    save_name : str
        save name for excel merged format

    Returns
    -------

    """
    index = ['GO_id', 'GO_name', 'depth', 'ref', 'aspect']
    tmp = pd.pivot_table(data, index=index, columns='sample_index',
                         aggfunc='first'
                         )
    tmp = tmp[['pvalue', 'enrichment_score', 'genes', 'n_genes', ]]
    if save_name:
        tmp.to_excel('{}.xlsx'.format(save_name),
                     merge_cells=True)
    return tmp


def write_table_to_html_with_figures(data, exp_data, save_name='index',
                                     out_dir=None, run_parallel=True):
    # create plots of everything
    if isinstance(data, str):
        data = pd.read_csv(data)

    fig_dict, to_remove = plot_genes_by_ont(
        data, save_name, out_dir, exp_data, run_parallel=run_parallel
    )

    for i in fig_dict:
        data.loc[data['GO_id'] == i, 'GO_name'] = fig_dict[i]

    data = data[~data['GO_id'].isin(to_remove)]

    tmp = pivot_table_for_export(data)

    html_out = save_name
    if out_dir is not None:
        html_out = os.path.join(out_dir, html_out)
    print("Saving to : {}".format(html_out))

    html_out = save_name + '_filter'
    if out_dir is not None:
        html_out = os.path.join(out_dir, html_out)

    write_filter_table(tmp, html_out)


class GoAnalysis(object):
    """
    Go analysis class.
    Uses the orangecontrib.bio package
    """

    def __init__(self, species='hsa', output_directory='tmp',
                 reference=None, verbose=False, experimental_data=None,
                 save_name='tmp', slim_name='goslim_pir'):

        self.exp_data = experimental_data
        self.data = None
        self.magine_go = MagineGO(species)
        self.savename = save_name
        self.out_dir = output_directory
        self.reference = reference
        self.verbose = verbose

        # options can be goslim_pir goslim_generic goslim_chembl
        self.slim_name = slim_name

        if reference is not None:
            self.gene_annotations = \
                set(reference).intersection(self.magine_go.gene_to_go.keys())
        else:
            self.gene_annotations = set(self.magine_go.gene_to_go.keys())
        self.number_of_total_reference_genes = len(self.gene_annotations)

        if not os.path.exists(self.out_dir):
            os.mkdir(self.out_dir)
        if not os.path.exists(os.path.join(self.out_dir, 'Figures')):
            os.mkdir(os.path.join(self.out_dir, 'Figures'))
        self.top_hits = []
        self.num_data_sets = 0
        self.created_go_pds = set()

    def _calculate_enrichment_single_sample(self, gene_list):
        """
        Performs enrichment analysis of list of genes

        Parameters
        ----------
        gene_list : array like
            List of genes to perform analysis
        """
        # checks first to see if all genes are annotated in GO
        genes_present = set()
        genes_missing = set()
        for i in gene_list:
            if type(i) != str:
                continue
            elif i in self.gene_annotations:
                genes_present.add(i)
            else:
                genes_missing.add(i)
                split_name = i.split(',')
                if len(split_name) > 1:
                    for n in split_name:
                        if n in self.gene_annotations:
                            genes_present.add(n)
        n_genes = len(genes_present)
        if len(genes_missing) > 0:
            print("Genes not in GO = {}".format(genes_missing))
        print("Number of genes given = {0}."
              " Number of genes in GO = {1}".format(len(gene_list), n_genes))

        # res = self.__calculate_enrichment(genes_present,
        # aspect=['F', 'P', 'C'])
        res = self.magine_go.calculate_enrichment(genes_present,
                                                  reference=self.reference)
        n = 0
        column_names = dict(GO_id=None, enrichment_score=None, pvalue=None,
                            genes=None, n_genes=None, aspect=None,
                            ref=None, depth=None, GO_name=None, slim=None)

        all_go_rows = []
        for go_id, (genes, p_value, ref) in res.items():
            n += 1
            go_row = column_names.copy()

            expected_value = n_genes * float(
                ref) / self.number_of_total_reference_genes
            enrichment = float(len(genes)) / expected_value

            go_row['GO_id'] = go_id
            go_row['enrichment_score'] = enrichment
            go_row['pvalue'] = p_value
            go_row['genes'] = list(np.sort(genes))
            go_row['n_genes'] = len(genes)
            go_row['aspect'] = self.magine_go.go_aspect[go_id]
            go_row['ref'] = len(self.magine_go.go_to_gene[go_id])
            go_row['depth'] = self.magine_go.go_depth[go_id]
            go_row['GO_name'] = self.magine_go.go_to_name[go_id]

            all_go_rows.append(go_row)
            if self.verbose:  # pragma: no cover
                print(go_id, float(len(genes) / float(ref)) * 100,
                      self.magine_go.go_to_name[go_id],
                      p_value, len(genes), ref,
                      enrichment)

        if n == 0:
            print("No significant p-values")

        return pd.DataFrame(all_go_rows, columns=column_names.keys())

    def calculate_enrichment(self, list_of_exp, labels=None):
        """
        Performs enrichment analysis of list of list of genes

        Parameters
        ----------
        list_of_exp: list_of_list or list
            Can be a single list of genes or a list of lists, where each sample
             is a list that contains species names.
        labels: list_like
            list of labels for each sample

        Returns
        -------
        pandas.DataFrame

        """
        self.num_data_sets = len(list_of_exp)

        if not all(isinstance(el, list) for el in list_of_exp):
            print("Running single GO enrichment since single list provided.")
            return self._calculate_enrichment_single_sample(list_of_exp)

        if self.num_data_sets == 0:
            print("Must provide at least one data list")
            print("Returning nothing")
            return

        if isinstance(labels, list):
            assert len(labels) == self.num_data_sets, \
                "If you provide a list, must be the same length" \
                " as list_of_exp"

        if labels is None:
            labels = range(0, self.num_data_sets)

        all_data = []
        for n, i in enumerate(list_of_exp):
            tmp = self._calculate_enrichment_single_sample(i)
            tmp['sample_index'] = labels[n]
            all_data.append(tmp)

        results = pd.concat(all_data)

        if results is None:  # pragma: no cover
            print('No data! Error! Returning nothing')
            return
        self.data = results
        return results

    def write_table_to_html(self, save_name='index', out_dir=None,
                            run_parallel=False):
        """
        Creates a table of all the plots of all genes for each GO term.
        
        Uses last calculated enrichment array.
        
        Parameters
        ----------
        save_name : str
            name of html output file
        out_dir : str, optional
            output path for all plots
        run_parallel : bool
            Create plots in parallel

        Returns
        -------

        """
        if out_dir is None:
            out_dir = self.out_dir
        print("BaseData", self.data)
        print("exp_data", self.exp_data)
        write_table_to_html_with_figures(
                self.data, self.exp_data, save_name, out_dir=out_dir,
                run_parallel=run_parallel
        )


class MagineGO(object):
    def __init__(self, species='hsa'):
        dirname = network_data_dir
        gene_to_go_name = os.path.join(dirname,
                                       '{}_gene_to_go.p'.format(species))
        go_to_gene_name = os.path.join(dirname,
                                       '{}_goids_to_genes.p'.format(
                                           species))
        go_to_go_name = os.path.join(dirname,
                                     '{}_goids_to_goname.p'.format(
                                         species))
        go_depth = os.path.join(dirname, '{}_godepth.p'.format(species))
        go_aspect = os.path.join(dirname, '{}_go_aspect.p'.format(species))
        for i in [gene_to_go_name, go_to_gene_name, go_to_go_name,
                  go_depth,
                  go_aspect]:
            if not os.path.exists(i):
                download_and_process_go(species=species)

        self.gene_to_go = pickle.load(open(gene_to_go_name, 'rb'))
        self.go_to_gene = pickle.load(open(go_to_gene_name, 'rb'))
        self.go_to_name = pickle.load(open(go_to_go_name, 'rb'))
        self.go_depth = pickle.load(open(go_depth, 'rb'))
        self.go_aspect = pickle.load(open(go_aspect, 'rb'))

    def calculate_enrichment(self, genes, reference=None,
                             evidence_codes=None,
                             aspect=None, use_fdr=True):
        """

        Parameters
        ----------
        genes : list
            list of genes
        reference : list
            reference list of species to calculate enrichment
        evidence_codes : list
            GO evidence codes
        use_fdr : bool
            Correct for multiple hypothesis testing

        Returns
        -------

        """

        # TODO check for alias for genes
        genes = set(genes)
        # TODO add aspects
        term_reference = self.go_to_gene.keys()
        aspect_dict = {
            'P': 'biological_process',
            'C': 'cellular_component',
            'F': 'molecular_function'
        }
        if aspect is None:
            term_reference = self.go_to_gene
            gene_reference = self.gene_to_go
        else:
            term_reference = dict()
            gene_reference = dict()

        if aspect is not None:
            for i in aspect:
                if i not in ['P', 'C', 'F']:
                    print("Error: Aspects are only 'P', 'C', and 'F' \n")
                    quit()
            for i in ['P', 'C', 'F']:
                if i in aspect:
                    term_reference = None

        # TODO add reference
        if reference:
            # TODO check for reference alias
            reference = set(reference)
            reference.intersection_update(set(self.gene_to_go.keys()))
        else:
            reference = set(self.gene_to_go.keys())

        # TODO add evidence_codes

        terms = set()
        for i in genes:
            if i in self.gene_to_go:
                for t in self.gene_to_go[i]:
                    terms.add(t)

        n_genes = len(genes)
        n_ref = float(len(reference))
        res = {}
        for term in terms:

            all_annotated_genes = set(self.go_to_gene[term])
            mapped_genes = genes.intersection(all_annotated_genes)
            n_mapped_genes = len(mapped_genes)

            if n_ref > len(all_annotated_genes):
                mapped_reference_genes = \
                    reference.intersection(all_annotated_genes)
            else:
                mapped_reference_genes = \
                    all_annotated_genes.intersection(reference)

            n_mapped_ref = len(mapped_reference_genes)

            prob = float(n_mapped_ref) / n_ref

            p_value = binom_test(n_mapped_genes, n_genes, prob, 'larger')

            res[term] = ([i for i in mapped_genes], p_value, n_mapped_ref)
        if use_fdr:
            res = sorted(res.items(), key=lambda x: x[1][1])
            fdr = fdrcorrection([p for _, (_, p, _) in res],
                                is_sorted=True)
            values = fdr[1]
            res = dict([(index, (genes, p, ref))
                        for (index, (genes, _, ref)), p in
                        zip(res, values)])
        return res

# All code below is old and will be removed. Keeping for just a bit longer.
'''  
# from magine.ontology.go_from_goatools import go as goa_tools
# from orangecontrib.bio import go
# from orangecontrib.bio.go import evidenceDict
from statsmodels.sandbox.stats.multicomp import fdrcorrection0
from statsmodels.stats.proportion import binom_test
from collections import defaultdict
    def __calculate_enrichment(self, genes, reference=None, evi_codes=None,
                               aspect=None, use_fdr=True, ):
        rev_genes_dict = self.annotations.get_gene_names_translator(genes)
        genes = set(rev_genes_dict.keys())
        if reference:
            ref_genes_dict = self.annotations.get_gene_names_translator(
                    reference)
            reference = set(ref_genes_dict.keys())
        else:
            reference = self.annotations.gene_names

        if aspect is None:
            aspects_set = set(["P", "C", "F"])
        elif isinstance(aspect, str):
            aspects_set = set([aspect])
        else:
            aspects_set = aspect

        evi_codes = set(evi_codes or evidenceDict.keys())

        annotations_dict = defaultdict(set)
        for gene in genes:
            for ann in self.annotations.gene_annotations[gene]:
                if ann.Evidence_Code in evi_codes and ann.Aspect in aspects_set:
                    annotations_dict[ann.GO_ID].add(ann)
        ref_annotations = set([ann for gene in reference for ann in
                               self.annotations.gene_annotations[gene] if
                               ann.Evidence_Code in evi_codes and ann.Aspect in aspects_set])

        terms = annotations_dict.keys()
        filtered_terms = [term for term in terms if
                          term in self.annotations.ontology]

        if len(terms) != len(filtered_terms):
            term_diff = set(terms) - set(filtered_terms)
            warnings.warn(
                    "%s terms in the annotations were not found in the "
                    "ontology." % ",".join(map(repr, term_diff)),
                    UserWarning)

        terms = self.annotations.ontology.extract_super_graph(filtered_terms)
        n_genes = len(genes)
        n_ref = float(len(reference))

        res = {}
        for i, term in enumerate(terms):
            # all_annotated_genes = self.annotations.get_all_genes(term)
            # 46 % of time is all_annotations
            all_annotations = self.annotations.get_all_annotations(
                term).intersection(ref_annotations)
            # 39.1 % of time is all_annotated_genes
            all_annotated_genes = set(
                    [ann.geneName for ann in all_annotations])

            mapped_genes = genes.intersection(all_annotated_genes)
            n_mapped_genes = len(mapped_genes)
            if len(reference) > len(all_annotated_genes):
                mapped_reference_genes = reference.intersection(
                        all_annotated_genes)
            else:
                mapped_reference_genes = all_annotated_genes.intersection(
                        reference)
            n_mapped_ref = len(mapped_reference_genes)
            prob = float(n_mapped_ref) / n_ref

            p_value = binom_test(n_mapped_genes, n_genes, prob, 'larger')

            res[term] = ([rev_genes_dict[g] for g in mapped_genes],
                         p_value, n_mapped_ref)
        if use_fdr:
            res = sorted(res.items(), key=lambda x: x[1][1])
            fdr = fdrcorrection0([p for _, (_, p, _) in res],
                                 is_sorted=True)
            values = fdr[1]
            res = dict([(index, (genes, p, ref))
                        for (index, (genes, _, ref)), p in
                        zip(res, values)])

        return res

    # TODO remove and replace with updated heatplots
    def __create_heatmaps(self, labels, savename):
        from magine.plotting.heatmaps import plot_heatmap, plot_heatmap_cluster
        # depracated
        data = self.data
        names, array = self.__sort_by_hierarchy(data)
        # creats plots of top and bottom of hierarchy sorted arrays
        plot_heatmap(array, names, labels, self.magine_go.go_to_name, start=0,
                     stop=100, savename='%s_top' % savename, )
        plot_heatmap(array, names, labels, self.magine_go.go_to_name,
                     start=-100, stop=None, savename='%s_bottom' % savename)
        tmp_array = array[:, :].copy()

        # Centroid clustering of data
        # Creates dendrogram plots
        clustered_array, clustered_names = plot_heatmap_cluster(
                tmp_array, names, savename="{}_dendrogram".format(savename),
                out_dir=self.out_dir)

        plot_heatmap(clustered_array, clustered_names, labels,
                     self.magine_go.go_to_name, start=0, stop=100,
                     savename='%s_top_clustered' % savename)
        plot_heatmap(clustered_array, clustered_names, labels,
                     self.magine_go.go_to_name, start=-100, stop=None,
                     savename='%s_bottom_clustered' % savename)

        figures = ['%s_top' % savename,
                   '%s_bottom' % savename,
                   '%s_clustered' % savename,
                   '%s_top_clustered' % savename,
                   '%s_bottom_clustered' % savename, ]
        html_pages = []

        for i in figures:
            html_pages.append(
                '<a href="Figures/{0}.pdf">{0}</a>'.format(i))

        self.html_pdfs = pd.DataFrame(html_pages, columns=['Clustered output'])
        self.__print_ranked_over_time(savename=savename, labels=labels)

    def __print_ranked_over_time(self, savename=None, labels=None, number=10,
                                 create_plots=True):
        """
        Order array by each sample and print information
        Parameters
        ----------
        savename : str
            name to save file as
        labels : list like
            list of samples
        number : int
            number of top hits to print
        create_plots : bool
            create a pdf of top hits

        Returns
        -------

        """
        from magine.plotting.heatmaps import plot_heatmap
        html_pages = []
        self.html_pdfs2 = []
        self.top_hits = []
        n = np.shape(self.data)[1] - 1
        score_names = []

        if labels is None:
            labels = list(range(n))
        for i in range(n):
            score_names.append('score_{0}'.format(i))

        for i in range(0, np.shape(self.data)[1] - 1):
            terms = self.__retrieve_top_ranked(i, number)
            tmp = []
            names = []
            for t in terms:
                names.append(t)
                tmp.append(np.array(
                    self.data_2[self.data_2['GO_id'] == t][score_names])[0])

            tmp = np.array(tmp)[::-1]
            names = np.array(names)[::-1]

            if create_plots:
                plot_heatmap(tmp, names, labels, start=-1 * number, stop=None,
                             savename='top_hits_entry_%i_%s' % (i, savename),)
                html_pages.append(
                    '<a href="Figures/top_hits_entry_{0}_{1}.pdf">{2}</a>'.format(
                        i, savename, labels[i]))
            terms_dict = {}
            for j in range(number):
                if j > len(tmp) - 1:
                    continue
                terms_dict[names[j]] = tmp[j]
                if tmp[j, i] == 0.0:
                    continue
                else:
                    print("Top hits = {0}, {1}, {2}, {3}".format(
                        i, names[j], self.global_go[names[j]], tmp[j, i]))

            self.top_hits.append(terms_dict)

        if create_plots:
            self.html_pdfs2 = pd.DataFrame(html_pages,
                                           columns=['Top hits per time'])

    def __retrieve_top_ranked(self, index, number=20):
        """
        Sorts array by column while excluding similar related GO terms
        Parameters
        ----------
        index : int
            column number to sort GO array by
        number : int
            number of terms wanted in top hits

        Returns
        -------

        """

        names, tmp = _sort_data_by_index(self.data, index)

        if len(names) < number:
            number = len(names)

        scores = {}
        for i in range(len(names)):
            scores[names[i]] = tmp[i, index]
        points = []
        counter = 0
        go_terms = list(names[-1 * number:])
        go_terms.reverse()
        # return go_terms

        term_to_add = -1 * number
        while counter < number:
            parents = dict(
                    [(term, self.__get_parents(term, go_terms)) for term in
                     go_terms])
            top_level_terms = [id for id in parents if not parents[id]]

            terms_to_remove = []
            index_to_add = []
            for term in top_level_terms:
                child = _get_children(term, go_terms, parents)
                if len(child) == 0:
                    counter += 1
                else:
                    term_to_add -= 1
                    if np.abs(term_to_add) > len(names):
                        continue
                    terms_to_remove.append(term)

                    index_to_add.append(names[term_to_add])
            for t in terms_to_remove:
                go_terms.remove(t)
            for t in index_to_add:
                go_terms.append(t)

        for j in range(number):
            points.append(go_terms[j])
        return points

    def __get_parents(self, term, data):
        """
        Get parents of GO term from provided lsit
        Parameters
        ----------
        term : str
            GO ontology term
        data : list
            list of GO terms in which ones wants to find parents of provided
            term

        Returns
        -------

        """
        parents = self.ontology.extract_super_graph([term])
        parents = [id for id in parents if id in data and id != term]
        lst = [set(self.ontology.extract_super_graph([id])) - set([id]) for id
               in parents]
        c = set.union(*lst) if lst else set()
        parents = [t for t in parents if t not in c]
        return parents

    def __sort_by_hierarchy(self, data):
        """
        Returns numpy array of GO terms by GO hierarchy
        Parameters
        ----------
        data : array
            Array of GO terms where first row is GO id, other rows enrichment
            values

        Returns
        -------

        """
        names = data[:, 0].copy()
        array = data[:, 1:].astype(np.float32).copy()
        parents = dict(
                [(term, self.__get_parents(term, names)) for term in names])
        top_level_terms = [id for id in parents if not parents[id]]
        term_list = []
        visited = []

        def _collect(go_term, parent):
            """
            Collects all children terms to a hierarchy
            Modified from bio.orange.go

            Parameters
            ----------
            go_term : str
                GO term
            parent : int or None
                Counter to provide GO hierarchy

            Returns
            -------

            """
            term_list.append((go_term, self.global_go[go_term], parent))
            parent = len(term_list) - 1
            for c in _get_children(go_term, names, parents):
                if c in visited:
                    continue
                else:
                    visited.append(c)
                    _collect(c, parent)

        for topTerm in top_level_terms:
            _collect(topTerm, None)
        term_list = np.array(term_list)
        list_sorted = list(term_list[:, 0])
        tmp_array = []
        for i in list_sorted:
            ind = np.where(names == i)
            tmp_array.append(ind[0][0])
        array = array[tmp_array]
        names = names[tmp_array]
        return names, array


def _get_children(term, data, parents):
    """

    Parameters
    ----------
    term : str
        GO term
    data : array

    parents :
        parents of terms

    Returns
    -------
    out : list
        list of children of GO term provided

    """
    return [id for id in data if term in parents[id]]


def _sort_data_by_index(data, index=0):
    """

    :param index:
    :param data:
    :return:
    """
    names = data[:, 0].copy()
    array = data[:, 1:].astype(np.float32).copy()
    step_size = (array[:, index]).argsort()
    names = names[step_size]
    array = array[step_size]
    return names, array
'''



"""
import os

import numpy as np
import pandas as pd
import pathos.multiprocessing as mp

from magine.networks.correlation_networks.calculate_correlation_network import \
    correlation_sampling


def test_correlation():

    dir_name = os.path.join(os.path.dirname(__file__),
                            'Data', 'jak_atra.csv.gz')

    data = pd.read_csv(dir_name)
    data = data[data['species_type'] == 'protein']
    tmp = pd.pivot_table(data, index=['protein'],
                         columns='time', aggfunc='first', fill_value=0.0)
    names = np.array(tmp.index.tolist())

    tmp = tmp['treated_control_fold_change']

    tmp = tmp.values
    good_index = (tmp == 0.0).sum(axis=1) < 3
    tmp = np.array(tmp)
    tmp = tmp[good_index]
    names = names[good_index]

    # tmp = tmp[:100]
    # names = names[:100]

    save_name = 'test'

    correlation_sampling(tmp, names, save_name, create_plots=True, pool=pool)


if __name__ == '__main__':
    pool = mp.Pool(4)
    test_correlation()
"""
